import requests
import logging

log = logging.getLogger(__name__)

from typer import Context, launch, echo, secho, Option, Typer, confirm, prompt

from kamocli.__main__ import app


@app.command("auth", help="Authenticate")
def cli(ctx: Context, api_key: str = Option(..., "-a", "--api-key", prompt="API Key")):
    domain = ctx.obj.session.profile.domain
    url = f"https://kamo.{domain}/.tenants"
    resp = None
    while not resp or resp.status_code == 403:
        if resp is not None:
            api_key = prompt("API Key incorrect. Please try again")
        echo(f"Testing {url} with API Key {api_key}...")
        resp = requests.get(url, timeout=2, headers={"Kamo-Api-Key": api_key})

    secho(f"API Key '{api_key}' has been added to current profile")
    ctx.obj.session.profile.api_key = api_key



import os, sys
import logging
import pathlib
import requests
import kamosdk
from kamosdk.session import ServiceSession
from kamosdk.exceptions import InvalidProfile
from kamocli.utils import dumps, out

from .utils import console_handler, abort

log = logging.getLogger(__name__)
help_string = """A utility for interfacing with Kamo dude"""

base_path = pathlib.Path(os.path.dirname(__file__))
commands_folder = base_path / "commands"


class ContextObj(object):
    def __init__(self):
        self.verbose = False
        self.profile_name = None
        self.home = os.getcwd()
        self.session = None
        self.service = None
        self.client = None
        self.base_path = base_path


from typer import Typer, Context, Argument, Option, echo, secho, confirm

app = Typer()


@app.callback(context_settings={"allow_extra_args": True, "ignore_unknown_options": True})
def cli(
    ctx: Context,
    verbose: str = Option(None, "-v", "--verbose", help="Verbose logging: info or debug"),
    profile: str = Option(None, "-p", "--profile", help="Profile name if you want to override the default profile"),
    output: str = Option("text", "-o", "--output", help="Output text or json"),
    tenant: str = Option(None, "-t", "--tenant", help="Override the tenant for this command"),
    domain: str = Option(None, "-d", "--domain", help="Override the domain for this command"),
):
    # by default we log out to console WARN and higher but can view info with -v
    # make sure this is called before importing the sdk so that we can grab all logging
    if verbose:
        console_handler.setLevel(getattr(logging, verbose.upper()))

    if ctx.obj is None:
        ctx.obj = ContextObj()
        ctx.obj.verbose = verbose
        ctx.obj.output = output
        try:
            ctx.obj.session = ServiceSession(profile=profile, tenant=tenant, domain=domain)
        except InvalidProfile as ex:
            log.debug(ex)
    try:
        out(
            f"Tenant: {ctx.obj.session.profile.tenant} - Profile: {ctx.obj.session.profile_name} - Endpoint: {ctx.obj.session.profile.domain}"
        )
    except:
        pass
    if ctx.invoked_subcommand not in ["tenants", "profile"]:
        if not ctx.obj.session:
            abort(f"You must create a profile with 'kamo profile add' to use this tool")

        if ctx.invoked_subcommand not in ["version"]:
            if not ctx.obj.session.profile.tenant:
                abort(f"You must select a tenant with 'kamo tenants' to use this tool")


@app.command()
def version(ctx: Context):
    """
    Show the versions of the CLI and all components
    """
    import kamocli

    ret = {"cli_version": kamocli.get_version_string()}
    ret["sdk_version"] = kamosdk.__version__
    ret["kamo_build_info"] = requests.get(ctx.obj.session.get_url("kamo")).json().get("build_info")
    ret["knode_build_info"] = requests.get(ctx.obj.session.get_url("knode")).json().get("build_info")

    if ctx.obj.output == "json" or 1:
        echo(dumps(ret))


from kamocli.commands import tenants, profile, login, db, nodes, builds, jobs, auth

app.add_typer(tenants.app, name="tenants", help="Manage tenants")
app.add_typer(profile.app, name="profile", help="Configure server profile to use")
app.add_typer(nodes.app, name="nodes", help="Manage EC2 instances")
app.add_typer(builds.app, name="builds")
app.add_typer(db.app, name="db")
app.add_typer(jobs.app, name="jobs")



"""
utils
~~~~~~~~~~
Various utilities for kamo-sdk functionality.
"""

import jwt
import logging
from urllib.parse import urlsplit
import requests
from requests import codes
import binascii
import os
from importlib.util import find_spec
import boto3
import uuid

from .exceptions import ServerError, InvalidToken

log = logging.getLogger(__name__)


def decode_token(token):
    try:
        decoded_token = jwt.decode(token, algorithms=["RS256"], verify=False)
        return decoded_token
    except (KeyError, jwt.InvalidTokenError) as ex:
        raise InvalidToken(f"Token could not be decoded ({ex}): {token}")


def check_resp_error(resp):
    response_json = None
    try:
        resp.raise_for_status()
    except Exception:
        desc = resp.text
        try:
            response_json = resp.json()
            desc = response_json["error"]
            log.info(response_json)
        except Exception:
            pass
        if resp.status_code == 403:
            desc = f"API Key invalid. Please run 'kamo auth' ({resp.text})"

        if not desc:
            desc = "Status code %s received (%s)" % (resp.status_code, resp.text)
        else:
            desc += " (code %s)" % resp.status_code
        if resp.status_code >= 500:
            desc = "Server error in call to %s" % resp.url
            desc += " - Response headers: %s" % resp.headers
            desc += " - Response body: %s" % resp.text
            log.error(desc)
        else:
            log.info("Server error in call to %s", resp.url)

        error = ServerError(desc, url=resp.url, response=response_json)
        raise error from None


def get_access_token(api_key, client_id, oauth_url):
    """ """
    log.info("Fetching access token from Cognito")
    client = boto3.client("cognito-idp")
    result = client.initiate_auth(
        AuthFlow="REFRESH_TOKEN_AUTH", AuthParameters={"REFRESH_TOKEN": api_key}, ClientId=client_id
    )
    return result["AuthenticationResult"]["AccessToken"]


def host_from_url(host):
    """
    Gets the raw URI host of a url with trailing slash, no matter how it is formatted.

    For example, www.server.com/something -> https://www.server.com/
                 http://www.server.com -> http://www.server.com/
                 http://localhost:8080/something -> http://localhost:8080/

    """
    if "://" not in host:
        host = f"https://{host}"
    parts = urlsplit(host)
    return "{}://{}/".format(parts.scheme, parts.netloc)


def generate_kamo_id(class_name: str) -> str:
    guid = str(uuid.uuid4())
    if "." in class_name or class_name.lower() != class_name:
        raise RuntimeError(f"{class_name} not a valid kamo class name")

    return class_name + "." + guid[:8]



import os


def read_version():
    directory = os.path.dirname(__file__)
    with open(os.path.join(directory, "VERSION"), "r") as version_file:
        version = version_file.readline().strip()
        return version


__version__ = read_version()


def get_version_string():
    return f"kamo-cli/{__version__}"



import click
import logging
import kamocli
import requests
import dateutil
import json
import time
from time import sleep
from kamocli.utils import dumps, abort, print_tab, fmt_date, check_tenant, success, out
from kamosdk.exceptions import ServerError
from tabulate import tabulate

log = logging.getLogger(__name__)

from typer import Typer, Context, Argument, Option, echo, secho, confirm

app = Typer()


@app.callback(
    help="Manage EC2 server instances",
    context_settings={"allow_extra_args": True, "ignore_unknown_options": True},
)
def cli(ctx: Context):
    pass


DEFAULT_INSTANCE_TYPE = "m5a.large"


@app.command("ensure")
@check_tenant
def ensure(
    ctx: Context,
    num: int = Argument(..., help="Number of instances"),
    instance_type: str = Option(DEFAULT_INSTANCE_TYPE, "--type", "-t", help="Type of EC2 instance to launch"),
    is_wait: bool = Option(False, "-w", "--wait", is_flag=True, help="Wait until instance is ready"),
):
    """
    Make sure that the specified number of instances are
    running on the tenant and terminate excessive instances if needed.
    Also cleans up any zombie entries in knode to make things nice and tidy.
    """
    MAX_NUM = 6
    if num > MAX_NUM:
        abort(f"You can ensure at most {MAX_NUM} instances")
    tenant = ctx.obj.session.profile.tenant
    nodes = find_nodes(ctx, tenant)
    if nodes:
        for node in nodes:
            url = ctx.obj.session.knode_url + f"/nodes/{node['node_id']}"
            if node["status"] == "unknown":
                out("Removing broken instance %s" % node["node_id"])
                try:
                    resp = ctx.obj.session.delete(url)
                    resp.raise_for_status()
                except Exception as e:
                    abort(
                        f"Could not remove broken instance {node['node_id']}. Please delete it manually before continuing."
                    )
        nodes = find_nodes(ctx, tenant)

    num_nodes_to_launch = num - len(nodes)
    if num_nodes_to_launch <= 0:
        txt = f"Tenant has {len(nodes)} running nodes and no nodes need to be launched\n"
        for node in nodes:
            txt += node["node_id"] + "\n"
        success(txt)

    url = ctx.obj.session.knode_url + "/nodes"
    data = {"instance_type": instance_type}
    tenant = ctx.obj.session.profile.tenant
    data["tenant"] = tenant
    out(f"Launching {num_nodes_to_launch} nodes for tenant {tenant}")
    new_nodes = []
    for i in range(num_nodes_to_launch):
        try:
            resp = ctx.obj.session.post(url, json=data)
            new_nodes.append(resp.json()["node_id"])
        except ServerError as ex:
            abort(ex.message)

    if is_wait:
        node_infos = wait_for_nodes(ctx, new_nodes)
        txt = "All nodes ready!\n"
        for node_id, node_info in node_infos.items():
            status = node_info["status"]
            txt += f"Node {node_id} is in state '{status}'\n"
        success(txt)
    else:
        success(f"{num_nodes_to_launch} nodes are being started")


@app.command("launch")
def launch(
    ctx: Context,
    instance_type: str = Option(DEFAULT_INSTANCE_TYPE, "--type", "-t", help="Type of EC2 instance to launch"),
    is_all: bool = Option(
        False,
        "-a",
        "--all",
        is_flag=True,
        help="Node should not have node affinity but be available to all tenants",
    ),
    is_wait: bool = Option(False, "-w", "--wait", is_flag=True, help="Wait until instance is ready"),
):
    """
    Launch a new EC2 instance
    """
    url = ctx.obj.session.knode_url + "/nodes"
    data = {"instance_type": instance_type}
    if not is_all:
        data["tenant"] = ctx.obj.session.profile.tenant
    try:
        resp = ctx.obj.session.post(url, json=data)
    except ServerError as ex:
        abort(ex.message)
    node_id = resp.json()["node_id"]
    if not ctx.obj.output == "json":
        secho(f"EC2 instance {node_id} is being launched...")
    if is_wait:
        node_info = wait_for_nodes(ctx, [node_id])[0]
        status = node_info["status"]
        success(f"Node {node_id} is in state '{status}'")


def wait_for_nodes(ctx, node_ids):
    start_time = time.time()
    url = ctx.obj.session.knode_url + "/nodes"
    ready_nodes = {}
    while len(node_ids) > len(ready_nodes):
        sleep(5.0)
        for node_id in node_ids:
            node_info = ctx.obj.session.get(url + f"/{node_id}").json()
            status = node_info["status"]
            if status != "launching":
                ready_nodes[node_id] = node_info
        diff = time.time() - start_time
        nodes = [n for n in node_ids if n not in ready_nodes]
        if nodes:
            txt = ", ".join(nodes)
            out(f"[{diff:.0f}s] Waiting for instances to become ready: {txt}")
    return ready_nodes


def find_nodes(ctx, tenant=None):
    resp = ctx.obj.session.get(ctx.obj.session.knode_url + "/nodes")
    resp.raise_for_status()
    content = resp.json()
    nodes = content["nodes"]
    if tenant:
        nodes = [n for n in content["nodes"] if n.get("tenant") == tenant]
    return nodes


@app.command("list")
@check_tenant
def list_nodes(
    ctx: Context,
    is_all: bool = Option(
        False,
        "-a",
        "--all",
        is_flag=True,
        help="View all nodes on the server, not just for my tenant",
    ),
):

    """
    List Nodes
    """
    tenant = ctx.obj.session.profile.tenant if not is_all else None
    nodes = find_nodes(ctx, tenant)

    if ctx.obj.output == "json":
        echo(dumps(nodes))
        return

    if not nodes:
        if is_all:
            abort("No nodes running on server")
        else:
            abort(
                f"No nodes running available to tenant '{tenant}'. Rerun this command with the --all flag to see all nodes."
            )

    node_list = []
    columns = ["node_id", "status", "registered", "last_modified", "instance_type", "tenant"]
    for node in nodes:
        node_row = []
        for k in columns:
            v = node.get(k)
            v = fmt_date(v)
            if k == "status":
                if v != "running":
                    v = click.style(v, fg="red")
            node_row.append(v)
        resources = node.get("resource_allocations")
        if resources:
            node_row.append(f"{resources['cpu']['current']}/{resources['cpu']['max']}")
            node_row.append(f"{resources['ram']['current']}/{resources['ram']['max']}")
            node_row.append(f"{resources['utilization']}%")
        node_list.append(node_row)
    columns.append("CPU")
    columns.append("RAM")
    columns.append("Utilization")
    echo(tabulate(node_list, headers=columns))


@app.command("terminate-all")
def terminate_all(
    ctx: Context,
    is_all: bool = Option(False, "-a", "--all", is_flag=True, help="Teminate all nodes for all tenants"),
    is_force: bool = Option(False, "-y", is_flag=True, help="Force a delete without a prompt"),
):
    """
    Terminate all nodes for the tenant
    """
    tenant = ctx.obj.session.profile.tenant if not is_all else None
    nodes = find_nodes(ctx, tenant)
    if not nodes:
        abort("No EC2 nodes found")
    if not is_force:
        if is_all:
            confirm(
                f"Are you sure you want to terminate {len(nodes)} EC2 instances for ALL tenants on the server?",
                abort=True,
            )
        else:
            confirm(
                f"Are you sure you want to terminate {len(nodes)} EC2 instances for the '{tenant}' tenant?",
                abort=True,
            )

    for node in nodes:
        url = ctx.obj.session.knode_url + f"/nodes/{node['node_id']}"

        resp = ctx.obj.session.delete(url)
        resp.raise_for_status()
    success(f"{len(nodes)} EC2 instances are being terminated")


@app.command("view")
def view(ctx: Context, node_id: str = Argument(..., help="EC2 Instance ID to view")):
    """
    View a specific EC2 instance
    """
    url = ctx.obj.session.knode_url + f"/nodes/{node_id}"

    try:
        resp = ctx.obj.session.get(url)
    except ServerError as ex:
        abort(ex)
    node = resp.json()
    if ctx.obj.output == "json":
        echo(dumps(node))
        return

    for k in [
        "node_id",
        "status",
        "last_modified",
        "registered",
        "instance_type",
        "availability_zone",
        "ami_id",
    ]:
        print_tab(k, node.get(k))
    resources = node.get("resource_allocations")
    if resources:
        print_tab("CPU Allocation", f"{resources['cpu']['current']}/{resources['cpu']['max']}")
        print_tab("RAM Allocation", f"{resources['ram']['current']}/{resources['ram']['max']}")
        print_tab("Utilization", f"{resources['utilization']}%")
    echo("")


@app.command()
def terminate(
    ctx: Context,
    node_id: str = Argument(..., help="EC2 Instance ID to terminate"),
    is_force: bool = Option(False, "-y", is_flag=True, help="Force a delete without a prompt"),
):
    """
    Terminate an EC2 Instance
    """
    url = ctx.obj.session.knode_url + f"/nodes/{node_id}"

    try:
        _ = ctx.obj.session.get(url)
    except ServerError as ex:
        abort(ex)

    if not is_force:
        confirm(
            f"Are you sure you want to terminate node {node_id}?",
            abort=True,
        )

    resp = ctx.obj.session.delete(url)
    resp.raise_for_status()
    echo(f"Node {node_id} has been terminated")



import logging
import requests
import dateutil
import json
from tabulate import tabulate
import tempfile
import collections

from kamosdk.exceptions import ServerError
from kamosdk.utils import generate_kamo_id
import kamocli
from kamocli.utils import (
    dumps,
    abort,
    print_tab,
    print_table,
    check_tenant,
    get_tenant_url,
    success,
    find_regions,
    find_builds,
    fmt_date,
    out,
)

log = logging.getLogger(__name__)

from typer import Typer, Context, Argument, Option, echo, secho, confirm, edit

app = Typer()


@app.callback(
    help="Kamo database management",
    context_settings={"allow_extra_args": True, "ignore_unknown_options": True},
)
def cli(ctx: Context):
    ctx.obj.tenant = ctx.obj.session.profile.tenant


@app.command("stats")
@check_tenant
def stats(ctx: Context):
    """
    View Kamo DB statistics
    """
    session = ctx.obj.session
    tenant = ctx.obj.tenant
    url = get_tenant_url(ctx.obj.session, ctx.obj.tenant)
    resp = session.get(url + "/db/stats")
    content = resp.json()
    if ctx.obj.output == "json":
        echo(dumps(content))
        return

    secho(f"\nStatistics for tenant {tenant}:\n", bold=True)
    lst = [list(content["counts"].values())]
    headers = list(content["counts"].keys())
    echo(tabulate(lst, headers=headers))

    if content["most_populated_regions"]:
        secho("\nMost populated regions:\n", bold=True)
        lst = []
        for region in content["most_populated_regions"]:
            lst.append([region["id"], region["objects"]])
        echo(tabulate(lst, headers=["region", "objects"]))

    if content["most_common_classes"]:
        secho("\nMost common classes:\n", bold=True)
        lst = []
        for instance in content["most_common_classes"]:
            lst.append([instance["kamo_class"], instance["instances"]])
        echo(tabulate(lst, headers=["Class", "Instances"]))
    echo("")


@app.command("clear")
def clear(
    ctx: Context,
    y: bool = Option(False, "-y", is_flag=True, help="Override 'Are you sure' prompt"),
):
    """
    Delete all kamo objects and regions from the database
    """
    session = ctx.obj.session
    url = get_tenant_url(ctx.obj.session, ctx.obj.tenant)
    if not y:
        confirm(
            f"Are you sure you want to delete the Kamo DB for the '{ctx.obj.tenant}' tenant?",
            abort=True,
        )
    session.delete(url + "/db")
    success(f"The database for '{ctx.obj.tenant}' has been cleared", fg="yellow", bold=True)


objects_app = Typer()
app.add_typer(objects_app, name="objects", help="Manage kamo objects")


@objects_app.command("list")
def objects(
    ctx: Context,
    class_name: str = Option("", "-c", "--class", help="Filter by object class"),
    kamo_id: str = Option("", "-k", "--kamo_id", help="filter on kamo_id"),
    region_id: str = Option("", "-r", "--region_id", help="filter on region_id"),
    num: int = Option(100, "-n", help="Maximum number of results"),
):
    """
    View list of all kamo objects in all regions
    """
    session = ctx.obj.session
    tenant = ctx.obj.tenant
    tenant_url = get_tenant_url(ctx.obj.session, ctx.obj.tenant)
    url = tenant_url + f"/db/objects?kamo_id={kamo_id}&class_name={class_name}&region_id={region_id}&end={num}"
    resp = session.get(url)
    content = resp.json()

    if ctx.obj.output == "json":
        echo(dumps(content))
        return

    lst = []
    for obj in content:
        lst.append([obj["id"], obj["region_id"]])

    if not lst:
        abort(f"No objects found for tenant {tenant}")

    headers = ["ID", "Region"]
    headers.extend(["Handler", "IP address"])
    secho(f"\nFound {len(lst)} objects for tenant {tenant}", bold=True)
    if len(lst) >= num:
        secho(f"Note that only the top {num} items are returned")
    echo(tabulate(lst, headers=headers))
    echo("")


@objects_app.command("report")
def report(
    ctx: Context,
):
    """
    Kamo item report for the current tenant
    """
    session = ctx.obj.session
    tenant = ctx.obj.tenant
    tenant_url = get_tenant_url(ctx.obj.session, ctx.obj.tenant)
    url = tenant_url + f"/db/objects?end=9999999"
    resp = session.get(url)
    content = resp.json()

    if ctx.obj.output == "json":
        echo(dumps(content))
        return

    items_by_class = collections.defaultdict(int)
    items_by_region = collections.defaultdict(int)
    for obj in content:
        class_name = obj["id"].split(".")[0]
        items_by_class[class_name] += 1
        items_by_region[obj["region_id"]] += 1

    secho("Items by region", bold=True)

    headers = ["Region", "Count"]
    lst = []
    for k, v in items_by_region.items():
        lst.append((k, v))
    echo(tabulate(lst, headers=headers))
    echo("")

    secho("Items by object class", bold=True)

    headers = ["Class", "Count"]
    lst = []
    for k, v in items_by_class.items():
        lst.append((k, v))
    echo(tabulate(sorted(lst), headers=headers))
    echo("")


@objects_app.command("cleanup")
def delete(
    ctx: Context,
    class_name: str,
    region_id: str = Option("", "-r", "--region_id", help="filter on region_id"),
    y: bool = Option(False, is_flag=True, help="Override 'Are you sure' prompt"),
):
    """
    Delete multiple kamo objects by class.

    TODO: Temporarily just call delete for each object. Needs to be refactored
    """
    session = ctx.obj.session
    tenant = ctx.obj.tenant
    tenant_url = get_tenant_url(ctx.obj.session, ctx.obj.tenant)
    url = tenant_url + f"/db/objects?class_name={class_name}&region_id={region_id}&end=99999"
    resp = session.get(url)
    content = resp.json()
    if not y:
        confirm(
            f"Are you sure you want to delete {len(content)} kamo objects in the '{tenant}' tenant?",
            abort=True,
        )

    for item in content:
        kamo_id = item["id"]
        url = tenant_url + f"/db/objects/{kamo_id}"
        out(f"Deleting item {kamo_id}")
        try:
            resp = session.delete(url)
        except ServerError as ex:
            abort(ex)
    success(f"{len(content)} items have been deleted from the '{tenant}' tenant")


object_app = Typer()
app.add_typer(object_app, name="object", help="Manage a single kamo object")


@object_app.callback(
    help="Object management",
    context_settings={"allow_extra_args": True, "ignore_unknown_options": True},
)
def cli(ctx: Context, kamo_id: str = Argument(..., help="Kamo ID of the object")):
    from urllib.parse import quote

    ctx.obj.kamo_id = quote(kamo_id)


@object_app.command("delete")
def delete(ctx: Context):
    """
    Delete the kamo object
    """
    kamo_id = ctx.obj.kamo_id
    session = ctx.obj.session
    tenant = ctx.obj.tenant
    tenant_url = get_tenant_url(ctx.obj.session, ctx.obj.tenant)
    url = tenant_url + f"/db/objects/{kamo_id}"
    try:
        resp = session.delete(url)
    except ServerError as ex:
        abort(ex)


regions_app = Typer()
app.add_typer(regions_app, name="regions", help="Manage kamo regions")


@object_app.command("view")
def view_object(ctx: Context):
    """
    View a kamo object
    """
    kamo_id = ctx.obj.kamo_id
    session = ctx.obj.session
    tenant = ctx.obj.tenant
    tenant_url = get_tenant_url(ctx.obj.session, ctx.obj.tenant)
    url = tenant_url + f"/db/objects/{kamo_id}"
    try:
        resp = session.get(url)
    except ServerError as ex:
        abort(ex)
    content = resp.json()
    if ctx.obj.output == "json":
        echo(dumps(content))
        return

    content["state"] = "(%s items)" % len(content.get("state", {}))
    print_table(content)

    region_id = content["region_id"]
    if region_id and region_id != kamo_id:
        out("\n")
        region_report(ctx, region_id)


@object_app.command("state")
def state_object(ctx: Context):
    """
    View a kamo object's state
    """
    kamo_id = ctx.obj.kamo_id
    session = ctx.obj.session
    tenant = ctx.obj.tenant
    tenant_url = get_tenant_url(ctx.obj.session, ctx.obj.tenant)
    url = tenant_url + f"/db/objects/{kamo_id}"
    try:
        resp = session.get(url)
    except ServerError as ex:
        abort(ex)
    content = resp.json()
    state = content.get("state", {})
    if ctx.obj.output == "json":
        echo(dumps(state))
        return

    secho(f"\nViewing state for object {kamo_id}\n", bold=True)
    for k, v in state.items():
        print_tab(k, v)


@object_app.command("edit")
def edit_object(ctx: Context, key: str = Option(None, "-k", "--key", help="Edit a specific key")):
    """
    Edit a kamo object state directly
    """
    kamo_id = ctx.obj.kamo_id
    session = ctx.obj.session
    tenant = ctx.obj.tenant
    tenant_url = get_tenant_url(ctx.obj.session, ctx.obj.tenant)
    url = tenant_url + f"/db/objects/{kamo_id}"
    try:
        resp = session.get(url)
    except ServerError as ex:
        abort(ex)
    state = resp.json()["state"]
    if key:
        js = state[key]
    else:
        js = state
    with tempfile.NamedTemporaryFile() as fp:
        txt = json.dumps(js, indent=4, default=str)
        while 1:
            out_txt = edit(txt)
            if not out_txt:
                abort("No changes were made")
            try:
                js = json.loads(out_txt)
                break
            except Exception as e:
                out(f"Json is invalid: {e}", fg="red")
                confirm("Would you like to open the editor again?", abort=True)
                txt = out_txt

        if key:
            state[key] = js
        else:
            state = js
        resp = session.put(url, json=state)
        if "message" in resp.json():
            success(resp.json()["message"])
        else:
            abort(str(resp.json()))


def region_report(ctx, region_id):
    out(f"Region {region_id}:")
    session = ctx.obj.session
    tenant = ctx.obj.tenant
    tenant_url = get_tenant_url(ctx.obj.session, ctx.obj.tenant)
    url = tenant_url + f"/db/regions/{region_id}"
    resp = session.get(url)
    content = resp.json()
    content["state"] = "(%s items)" % len(content.get("state", {}))
    print_table(content)
    handler_url = content["handler_url"]
    if handler_url:
        try:
            resp = session.get(handler_url)
            content = resp.json()
            server_id = content["id"]
            secho(f"\nHandler {server_id}:")
            content["state"] = "(%s items)" % len(content.get("state", {}))
            print_table(content)
        except ServerError:
            out(f"No Unreal server is currently handling region {region_id}")


regions_app = Typer()
app.add_typer(regions_app, name="regions", help="Manage kamo regions")


@regions_app.command("list")
@check_tenant
def list_regions(ctx: Context):
    """
    List Kamo regions
    """
    session = ctx.obj.session
    tenant = ctx.obj.tenant
    url = get_tenant_url(ctx.obj.session, ctx.obj.tenant)
    resp = session.get(url + "/db/regions?start=0&end=100")
    content = resp.json()
    if ctx.obj.output == "json":
        echo(dumps(content))
        return
    secho(f"Tenant {ctx.obj.tenant} is handling the following regions:")
    lst = []
    for region in resp.json()["regions"]:
        lst.append([region["id"], region["handler_id"]])

    if not lst:
        abort(f"No regions found for tenant {tenant}")

    echo(f"Regions for tenant {tenant}:")
    echo(tabulate(lst, headers=["ID", "Handler"]))


@regions_app.command("create")
@check_tenant
def create_region(
    ctx: Context,
    region: str = Argument(..., help="Kamo ID of region to create"),
    state: str = Option('{"handler": ""}', "-s", "--state", show_default=True, help="Region initial state"),
):
    """
    Create a new kamo region. Use a comma-separated list to create multiple identical regions
    """
    session = ctx.obj.session
    tenant = ctx.obj.tenant
    url = get_tenant_url(ctx.obj.session, tenant)
    state = json.loads(state)
    regions = [r.strip() for r in region.split(",")]
    result = ""
    for region in regions:
        data = {"id": region, "state": state}
        try:
            resp = session.post(url + f"/db/regions/", json=data)
        except ServerError as ex:
            abort(ex)
        content = resp.json()
        result += f"\nRegion '{region}' has been created in tenant '{tenant}'"
    success(result)


@regions_app.command("delete-all")
@check_tenant
def delete_all_regions(
    ctx: Context,
    y: bool = Option(False, is_flag=True, help="Override 'Are you sure' prompt"),
):
    """
    Delete all kamo regions
    """
    session = ctx.obj.session
    tenant = ctx.obj.tenant
    url = get_tenant_url(ctx.obj.session, tenant)
    regions = find_regions(ctx)
    if not regions:
        abort(f"Tenant '{tenant}' has no regions")
    if not y:
        confirm(
            f"Are you sure you want to delete {len(regions)} regions in the '{tenant}' tenant?",
            abort=True,
        )

    for region in regions:
        _ = session.delete(url + f"/db/regions/{region}")
    success(f"{len(regions)} have been deleted in tenant '{tenant}'")


region_app = Typer()
app.add_typer(region_app, name="region", help="Manage a single kamo region")


@region_app.callback(
    help="Region management",
    context_settings={"allow_extra_args": True, "ignore_unknown_options": True},
)
def cli(ctx: Context, region: str = Argument(..., help="Kamo ID of region to view")):
    ctx.obj.region = region


@region_app.command("view")
def view_region(ctx: Context):
    """
    View a single kamo region
    """
    region = ctx.obj.region
    session = ctx.obj.session
    url = get_tenant_url(ctx.obj.session, ctx.obj.tenant)
    resp = session.get(url + f"/db/regions/{region}")
    content = resp.json()
    if ctx.obj.output == "json":
        echo(dumps(content))
        return
    print_table(content)


@region_app.command("delete")
def delete_region(
    ctx: Context,
    y: bool = Option(False, is_flag=True, help="Override 'Are you sure' prompt"),
):
    """
    Delete the selected kamo region
    """
    region = ctx.obj.region
    session = ctx.obj.session
    tenant = ctx.obj.tenant
    url = get_tenant_url(ctx.obj.session, tenant)
    try:
        _ = session.get(url + f"/db/regions/{region}")
    except ServerError as ex:
        abort(ex)
    if not y:
        confirm(
            f"Are you sure you want to delete region {region} from the '{ctx.obj.tenant}' tenant?",
            abort=True,
        )

    resp = session.delete(url + f"/db/regions/{region}")
    success(f"Region '{region}' has been deleted in tenant '{tenant}'")


@region_app.command("create-object")
def region_create_object(
    ctx: Context,
    class_name: str,
    kamo_id: str = Option(None, "-k", "--kamo-id", help="Unique kamoID (optional)"),
    state: str = Option("{}", "-s", "--state", help="Initial kamo state for the object"),
):
    """
    Create an object in the region
    """
    region = ctx.obj.region
    session = ctx.obj.session
    kamo_id = kamo_id or generate_kamo_id(class_name)
    url = get_tenant_url(ctx.obj.session, ctx.obj.tenant)
    data = {"id": kamo_id, "region_id": region, "state": json.loads(state)}
    try:
        resp = session.post(f"{url}/db/objects", json=data)
    except ServerError as ex:
        abort(ex)
    content = resp.json()
    if ctx.obj.output == "json":
        echo(dumps(content))
        return
    success(f"Kamo object {kamo_id} has been created in region {region}")


@region_app.command("delete-object")
def region_delete_object(ctx: Context, kamo_id: str):
    """
    delete a specific object in the region
    """
    region = ctx.obj.region
    session = ctx.obj.session
    url = get_tenant_url(ctx.obj.session, ctx.obj.tenant)
    resp = session.delete(f"{url}/db/objects/{kamo_id}?region_id={region}")
    content = resp.json()
    if ctx.obj.output == "json":
        echo(dumps(content))
        return
    success(f"Kamo object {kamo_id} has been deleted in region {region}")


@region_app.command("delete-object-class")
def region_delete_objects_class(ctx: Context, class_name: str):
    """
    delete all object by class in the region
    """
    region = ctx.obj.region
    session = ctx.obj.session
    url = get_tenant_url(ctx.obj.session, ctx.obj.tenant)
    resp = session.delete(f"{url}/db/objects/?class_name={class_name}&region_id={region}")
    content = resp.json()
    if ctx.obj.output == "json":
        echo(dumps(content))
        return
    msg = content.get("message")
    success(f"Kamo objects with class {class_name} have been deleted in region {region}. Server response: {msg}")


@region_app.command("objects")
def region_objects(ctx: Context, class_name: str = Option("", "-c", "--class", help="Filter by object class")):
    """
    View a list of objects in the region
    """
    region = ctx.obj.region
    session = ctx.obj.session
    url = get_tenant_url(ctx.obj.session, ctx.obj.tenant)
    resp = session.get(f"{url}/db/objects?start=0&end=1000&region_id={region}&class_name={class_name}")
    content = resp.json()
    if ctx.obj.output == "json":
        echo(dumps(content))
        return
    objects = content
    lst = []
    for obj in objects:
        lst.append([obj["id"], obj["region_id"]])
    echo(tabulate(lst, headers=["kamo_id", "region_id"]))
    echo("")


@app.command("handlers")
def handlers(ctx: Context):
    """
    View list of all kamo handlers
    """
    session = ctx.obj.session
    tenant = ctx.obj.tenant
    url = get_tenant_url(ctx.obj.session, ctx.obj.tenant)
    resp = session.get(url + "/db/handlers?start=0&end=100&full_state=true")
    content = resp.json()
    if ctx.obj.output == "json":
        echo(dumps(content))
        return
    secho(f"Tenant {ctx.obj.tenant} has the following handlers:")
    lst = []
    for handler in resp.json()["handlers"]:
        state = handler["state"]
        handler_row = [handler["id"], state["ip_address"] + ":" + str(state["port"])]
        handler_row.append(dateutil.parser.parse(state["start_time"]).strftime("%Y-%m-%d %H:%M"))
        handler_row.append(dateutil.parser.parse(state["last_refresh"]).strftime("%Y-%m-%d %H:%M"))
        lst.append(handler_row)

    if not lst:
        abort(f"No handlers found for tenant {tenant}")

    echo(f"Handlers for tenant {tenant}:")

    echo(tabulate(lst, headers=["ID", "IP Address", "Created", "Updated"]))


@regions_app.command("clear-handlers")
def clear_handlers(
    ctx: Context,
    y: bool = Option(False, "-y", is_flag=True, help="Override 'Are you sure' prompt"),
):
    """
    Clear all handlers from all regions (debug only!)
    """
    session = ctx.obj.session
    tenant = ctx.obj.tenant
    url = get_tenant_url(ctx.obj.session, ctx.obj.tenant)
    regions = find_regions(ctx)
    if not y:
        confirm(
            f"Are you sure you want to clear handlers from {len(regions)} regions in the '{ctx.obj.tenant}' tenant?",
            abort=True,
        )
    for region in regions:
        resp = session.get(url + f"/db/regions/{region}")
        content = resp.json()
        state = content["state"]
        url = content["url"]
        state["handler"] = ""
        _ = session.put(url + "?force=true", json=state)

    success(f"All regions for tenant {tenant} now have no handler")


snapshots_app = Typer()

app.add_typer(snapshots_app, name="snapshots", help="Manage database snapshots")


@snapshots_app.command("list")
def list_snapshots(ctx: Context):
    """
    Get a list of snapshots for tenant
    """
    session = ctx.obj.session
    tenant = ctx.obj.tenant
    url = get_tenant_url(ctx.obj.session, ctx.obj.tenant)
    resp = session.get(url + "/db/snapshots")
    if ctx.obj.output == "json":
        echo(resp.json())
        return

    lst = []
    headers = ["Name", "Timestamp", "Regions", "Objects"]
    for snapshot in resp.json()["snapshots"]:
        lst.append(
            [
                snapshot["name"],
                fmt_date(snapshot["timestamp"]),
                snapshot["counts"].get("regions"),
                snapshot["counts"].get("objects"),
            ]
        )

    echo(f"\nSnapshots available for tenant '{tenant}:")
    echo(tabulate(lst, headers=headers))


@snapshots_app.command("delete")
def delete_snapshot(
    ctx: Context,
    name: str = Argument(..., help="Name of the snapshot to delete"),
    y: bool = Option(False, is_flag=True, help="Override 'Are you sure' prompt"),
):
    """
    Delete a snapshot for tenant
    """
    session = ctx.obj.session
    tenant = ctx.obj.tenant
    url = get_tenant_url(ctx.obj.session, ctx.obj.tenant)
    if not y:
        confirm(
            f"Are you sure you want to delete snapshot '{name}' for the '{ctx.obj.tenant}' tenant?",
            abort=True,
        )
    try:
        _ = session.get(url + f"/db/snapshots/{name}")
        resp = session.delete(url + f"/db/snapshots/{name}")
    except ServerError as ex:
        abort(ex)
    success(f"Snapshot '{name}' has been deleted from tenant '{tenant}'")


@snapshots_app.command("create")
def create_snapshot(ctx: Context, name: str = Argument(..., help="Name of the snapshot to create")):
    """
    Create a snapshot for tenant from current state
    """
    session = ctx.obj.session
    tenant = ctx.obj.tenant
    url = get_tenant_url(ctx.obj.session, ctx.obj.tenant)
    data = {"name": name}
    try:
        resp = session.post(url + f"/db/snapshots", json=data)
    except ServerError as ex:
        abort(ex)
        print_table(resp.json())
    success(f"Snapshot '{name}' has been created for '{tenant}'")


@snapshots_app.command("restore")
def restore_snapshot(
    ctx: Context,
    name: str = Argument(..., help="Name of the snapshot to restore"),
    y: bool = Option(False, is_flag=True, help="Override 'Are you sure' prompt"),
):
    """
    Restore a snapshot to the current tenant, overriding the current database
    """
    session = ctx.obj.session
    tenant = ctx.obj.tenant
    url = get_tenant_url(ctx.obj.session, ctx.obj.tenant)
    if not y:
        confirm(
            f"Are you sure you want to restore snapshot '{name}' for the '{ctx.obj.tenant}' tenant?",
            abort=True,
        )

    try:
        _ = session.get(url + f"/db/snapshots/{name}")
        resp = session.post(url + f"/db/snapshots/{name}/restore")
    except ServerError as ex:
        abort(ex)
    success(f"Snapshot '{name}' has been restored for '{tenant}'")


@snapshots_app.command("delete-all")
def delete_all_snapshots(
    ctx: Context,
    y: bool = Option("...", is_flag=True, help="Override 'Are you sure' prompt"),
):
    """
    Delete all snapshots from tenant
    """
    session = ctx.obj.session
    tenant = ctx.obj.tenant
    url = get_tenant_url(ctx.obj.session, ctx.obj.tenant)
    resp = session.get(url + "/db/snapshots")
    snapshots = resp.json()["snapshots"]
    if not y:
        confirm(
            f"Are you sure you want to delete {len(snapshots)} snapshots for the '{ctx.obj.tenant}' tenant?",
            abort=True,
        )
    for snapshot in snapshots:
        name = snapshot["name"]
        try:
            resp = session.delete(url + f"/db/snapshots/{name}")
        except ServerError as ex:
            abort(ex)
    success(f"All {len(snapshots)} snapshots have been deleted from tenant '{tenant}'")



import click
import logging
import kamocli
import requests

from kamocli.utils import dumps, abort, success, print_table
from kamosdk.exceptions import ServerError

log = logging.getLogger(__name__)

from typer import Typer, Context, Argument, Option, echo, secho, confirm

app = Typer()


@app.command("list")
def list_tenants(ctx: Context):
    """
    List available tenants on the server
    """
    resp = ctx.obj.session.get(ctx.obj.session.kamo_url + "/.tenants")
    resp.raise_for_status()
    content = resp.json()
    if ctx.obj.output == "json":
        echo(dumps(content))
        return

    if not content.get("tenants"):
        abort("No tenants found")

    tenant = ctx.obj.session.profile.tenant or "(none)"
    secho(f"Currently selected tenant: " + click.style(tenant, bold=True) + "\nAvailable tenants:")
    for tenant in content["tenants"]:
        print_table(tenant)
        echo("")


@app.command("set")
def set_tenant(ctx: Context, tenant_name):
    """
    Pick a tenant to use for tenant-specific commands
    """
    tenant_name = tenant_name.lower()
    if not ctx.obj.session:
        abort("No profile found. Please add a profile with kamo profile add")
    resp = ctx.obj.session.get(ctx.obj.session.kamo_url + "/.tenants")
    resp.raise_for_status()
    content = resp.json()
    if not content.get("tenants"):
        abort("No tenants found on server")

    tenants = set()
    for t in content["tenants"]:
        tenants.add(t["name"].lower())
    if tenant_name not in tenants:
        abort("Invalid tenant. Please choose from: {}".format(", ".join(tenants)))

    ctx.obj.session.profile.tenant = tenant_name
    success("Tenant has been set to: {}".format(tenant_name), bold=True)


@app.command("create")
def create_tenant(ctx: Context, tenant_name: str):
    """
    Create a new tenant
    """
    tenant_name = tenant_name.lower()
    try:
        _ = ctx.obj.session.get(ctx.obj.session.kamo_url + f"/.tenants/{tenant_name}")
        abort(f"Tenant {tenant_name} already exists")
    except ServerError:
        pass
    data = {"name": tenant_name}

    _ = ctx.obj.session.post(ctx.obj.session.kamo_url + "/.tenants", json=data)
    success(f"Tenant {tenant_name} has been created.", bold=True)


@app.command("delete")
def delete_tenant(
    ctx: Context,
    tenant_name: str,
    is_force: bool = Option(False, "-y", is_flag=True, help="Delete the tenant without a prompt"),
):
    """
    Delete a tenant
    """
    tenant_name = tenant_name.lower()
    try:
        _ = ctx.obj.session.get(ctx.obj.session.kamo_url + f"/.tenants/{tenant_name}")
        abort(f"Tenant {tenant_name} already exists")
    except ServerError:
        pass
    if not is_force:
        confirm(
            f"Are you sure you want to permanently delete the tenant '{tenant_name}'?",
            abort=True,
        )

    _ = ctx.obj.session.delete(ctx.obj.session.kamo_url + f"/.tenants/{tenant_name}")
    success(f"Tenant {tenant_name} has been permanently deleted.", bold=True)



"""
kamosdk
~~~~~~~~~~
kamosdk is a python module for interfacing with Kamo Services.

"""
import os
import logging

from .config import Config

log = logging.getLogger()

cfg = Config()


def read_version():
    directory = os.path.dirname(__file__)
    with open(os.path.join(directory, "VERSION"), "r") as version_file:
        version = version_file.readline().strip()
        return version


__version__ = read_version()



import time
import sys
import kamocli
import os
import requests
import dateutil
from urllib.parse import urlparse, urljoin
from loguru import logger
import boto3
from io import BytesIO
import json
import botocore

from kamocli.utils import (
    dumps,
    abort,
    print_tab,
    fmt_date,
    check_tenant,
    print_table,
    out,
    success,
    get_builds_from_s3_url,
)
from kamosdk.exceptions import ServerError
from tabulate import tabulate

from typer import Typer, Context, Argument, Option, echo, secho, confirm

app = Typer()


@app.callback(
    help="Manage UE4 server builds",
    context_settings={"allow_extra_args": True, "ignore_unknown_options": True},
)
def cli(ctx: Context):
    ctx.obj.tenant = ctx.obj.session.profile.tenant
    ctx.obj.builds_path = ctx.obj.session.profile.builds_path
    if not ctx.obj.builds_path and ctx.invoked_subcommand != "path":
        abort("Please set s3 build path with 'kamo builds path'")


@app.command("path")
def set_builds_path(
    ctx: Context,
    url: str = Argument(
        ..., help="Full s3 url to builds. e.g. s3://at-teamcity-artifacts/Existence/Existence_BuildServer/"
    ),
):
    if not url.startswith("s3://"):
        abort("S3 url must start with 's3://'")
    url = os.path.join(url, "")  # add trailing / if needed

    builds = get_builds_from_s3_url(url, num=1)
    if not builds:
        secho(f"No builds found in {url}.", fg="red")
        echo(f"This tool looks for subfolders containing 'buildinfo.json' which must at the very least have a 'version' key")
    else:
        echo(f"Builds found in path '{url}'. Adding to profile")
    ctx.obj.session.profile.builds_path = url


@app.command("list")
def list_builds(ctx: Context, num: int = Option(10, help="Number of the most recent builds to view")):
    """List server builds"""
    tenant = ctx.obj.session.profile.tenant
    builds = list(get_builds_from_s3_url(ctx.obj.builds_path, num).values())
    if ctx.obj.output == "json":
        echo(dumps(builds))
        return
    lst = []
    headers = ["Version", "Timestamp", "Branch", "S3 Path"]
    for build in builds:
        lst.append([build["version"], fmt_date(build.get("timestamp")), build.get("branch"), build["s3_path"]])
    echo(tabulate(lst, headers=headers))


@app.command("view")
def view_build(
    ctx: Context,
    version: str = Argument(..., help="Build version to view"),
    is_wait: bool = Option(
        False, "-w", "--wait", is_flag=True, help="Wait for the build to be ready if it is not found"
    ),
):

    """View details about a specific build"""
    tenant = ctx.obj.session.profile.tenant
    url = ctx.obj.session.knode_url + f"/builds/{tenant}/{version}"
    found = False

    if is_wait:
        start_time = time.time()
        while time.time() < start_time + 600.0:
            try:
                resp = ctx.obj.session.get(url)
                found = True
                break
            except ServerError as ex:
                out(f"Waiting for build {version}")
                time.sleep(10.0)
        if found:
            out(f"Build {version} is available", bold=True)
        else:
            out(f"Timeout waiting for build {version} to become available", fg="red")
            sys.exit(1)
    else:
        try:
            resp = ctx.obj.session.get(url)
        except ServerError as ex:
            abort(ex)

    content = resp.json()
    if ctx.obj.output == "json":
        print(dumps(content))
        return
    print_table(content["manifest"][0])


@app.command("delete")
def delete_build(
    ctx: Context,
    version: str = Argument(..., help="Build version to delete"),
    is_force: bool = Option(False, "-y", is_flag=True, help="Force a delete without a prompt"),
):
    """Delete a build"""
    tenant = ctx.obj.session.profile.tenant
    url = ctx.obj.session.knode_url + f"/builds/{tenant}/{version}"
    if not is_force:
        confirm(
            f"Are you sure you want to delete build '{version}' for the '{tenant}' tenant?",
            abort=True,
        )

    try:
        _ = ctx.obj.session.delete(url)
    except ServerError as ex:
        abort(ex)

    secho(f"Build {version} has been deleted for tenant {tenant}", bold=True)



#!/usr/bin/env python

import botocore.session
import click
import typer
import configparser
import logging
import os
import sys
import json
from jsonpath_rw import parse
import yaml
import dateutil
from functools import wraps
from kamosdk.exceptions import ServerError
from urllib.parse import urlparse, urljoin
import boto3
from io import BytesIO
import json
import botocore
from loguru import logger
from operator import attrgetter, itemgetter

log_formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")

console_handler = logging.StreamHandler()
console_handler.setFormatter(log_formatter)
console_handler.setLevel(logging.WARN)

root_logger = logging.getLogger()
root_logger.setLevel(logging.DEBUG)
root_logger.addHandler(console_handler)

log = logging.getLogger(__name__)


def print_error(*txt):
    color_print(*txt, color="red")


def print_warn(*txt):
    color_print(*txt, color="yellow")


def print_success(*txt):
    color_print(*txt, color="green")


def color_print(*args, **kwargs):
    color = kwargs.get("color", "reset")
    for x in args:
        typer.echo(typer.style(x, fg=color))


def print_tab(key, val):
    val_txt = val
    if isinstance(val, dict):
        val_txt = ", ".join([f"{k}={v}" for k, v in val.items()])
    val = typer.style(str(val_txt), bold=True)
    typer.echo("{0:20}{1}".format(key, val))


def print_table(obj):
    try:
        for k, v in obj.items():
            print_tab(k, v)
    except Exception:
        typer.echo(obj)


def dumps(obj):
    try:
        return json.dumps(obj, indent=4, sort_keys=True)
    except Exception:
        return repr(obj)


def output_pretty_json(dct, keys):
    ret = ""
    for k in keys:
        jsonpath_expr = parse(k)
        ret = [match.value for match in jsonpath_expr.find(dct)]
        lst = [""]
        if ret:
            if isinstance(ret[0], list) or isinstance(ret[0], dict):
                lst = yaml.dump(ret[0]).split("\n")
            else:
                lst = [str(ret[0])]
        print_tab(k.split(".")[-1], lst[0])
        if len(lst) > 1:
            for v in lst[1:]:
                print_tab("", v)


def find_aws_credentials(profile):
    """
    Returns the aws credentials for the specified profile.
    If no profile is passed in, returns the credentials for the currently selected profile

    Args:
        profile name

    Returns:
        Dict containing at least aws_access_key_id, aws_secret_access_key

    Raises:
        RuntimeError is no default profile or the named profile was not found

    """
    if not profile:
        access_key = None
        secret_key = None
        token = ""
        credentials = botocore.session.get_session().get_credentials()
        if credentials:
            access_key = credentials.access_key
            secret_key = credentials.secret_key
            token = getattr(credentials, "token") or ""
        if not access_key or not secret_key:
            raise RuntimeError("No Default AWS profile set")

        return {
            "aws_access_key_id": access_key,
            "aws_secret_access_key": secret_key,
            "aws_session_token": token,
        }

    folder = os.path.join(os.path.expanduser("~"), ".aws")
    filename = os.path.join(folder, "credentials")
    cfg = configparser.ConfigParser()
    with open(filename) as fp:
        cfg.readfp(fp)
        ret = {}
        if profile not in cfg:
            raise RuntimeError("No AWS profile '%s' found in %s" % (profile, filename))
        for key in cfg[profile]:
            ret[key] = cfg[profile][key]
    return ret


def abort(reason, code=1):
    """
    exit with non-zero exit code and write reason for error to stderr.
    If we are outside a typer context and exception will be raised instead
    """
    ctx = click.get_current_context(silent=True)
    if not ctx:
        raise Exception(f"Abnormal Termination: {reason}")

    if ctx.obj and ctx.obj.output == "json":
        ret = {"success": False, "exit_code": code, "message": str(reason)}
        typer.echo(json.dumps(ret))
    else:
        typer.echo(typer.style(str(reason), fg="red"), file=sys.stderr)
    sys.exit(code)


def out(txt, **kw):
    ctx = click.get_current_context(silent=True)
    if ctx and ctx.obj and ctx.obj.output == "json":
        log.info(txt)
    else:
        typer.secho(txt, **kw)


def success(reason, response=None, **kw):
    ctx = click.get_current_context(silent=True)
    if not ctx:
        print(reason)

    exit_code = 0
    if ctx.obj and ctx.obj.output == "json":
        ret = {
            "success": True,
            "exit_code": exit_code,
            "message": str(reason),
            "response": response,
        }
        typer.echo(json.dumps(ret))
    else:
        if not kw:
            kw = {"fg": "white", "bold": True}
        typer.secho(str(reason), **kw)
    sys.exit(exit_code)


def check_profile(ctx):
    if not ctx.obj.client:
        abort("You have no active profile or selected profile is invalid. Run 'kamo profile add [name]'")


def fmt_date(dt: str):
    try:
        return dateutil.parser.parse(dt).strftime("%Y-%m-%d %H:%M")
    except:
        return dt


def check_tenant(func):
    @wraps(func)
    def _check(*args, **kwargs):
        ctx = click.get_current_context()
        tenant = ctx.obj.session.profile.tenant
        if not tenant:
            abort("Before using this command please set a tenant by running: kamo tenants set [name]")
        else:
            return func(*args, **kwargs)

    return _check


def get_tenant_url(session, tenant):
    url = session.kamo_url + f"/{tenant}"
    return url


def find_regions(ctx):
    url = get_tenant_url(ctx.obj.session, ctx.obj.tenant)
    resp = ctx.obj.session.get(url + "/db/regions?start=0&end=100")
    ret = set()
    for region in resp.json()["regions"]:
        ret.add(region["id"])
    return ret


def find_builds(ctx):
    tenant = ctx.obj.session.profile.tenant
    url = ctx.obj.session.knode_url + f"/builds/{tenant}"
    resp = ctx.obj.session.get(url)
    ret = []
    for build in resp.json()["builds"]:
        ret.append(build["version"])
    return ret


def find_nodes(ctx):
    url = ctx.obj.session.knode_url + f"/nodes"
    resp = ctx.obj.session.get(url)
    ret = {}
    for node in resp.json()["nodes"]:
        ret[node["node_id"]] = node
    return ret


def format_region_name(map_name: str, volume_name: str = None, instance_id: str = None):
    region_name = "region." + map_name
    if volume_name:
        region_name += "^" + volume_name
    if instance_id:
        instance_id += "#" + instance_id
    return region_name.lower()


def parse_region_name(region_name):
    volume_name = instance_id = None
    region_name = region_name.replace("region.", "")
    if "#" in region_name:
        region_name, instance_id = region_name.split("#")
    if "^" in region_name:
        region_name, volume_name = region_name.split("^")
    return region_name, volume_name, instance_id


def get_builds_from_s3_url(url, num=10):
    buildinfo_filename = "buildinfo.json"
    parts = urlparse(url)
    bucket_name = parts.netloc
    path = parts.path[1:]
    logger.debug("Finding builds in bucket {}, path {}", bucket_name, path)

    client = boto3.client("s3")
    s3 = boto3.resource("s3")
    paginator = client.get_paginator("list_objects")
    paths = []
    try:
        for result in paginator.paginate(Bucket=bucket_name, Prefix=path, Delimiter="/"):
            folders = [f["Prefix"] for f in result.get("CommonPrefixes", [])]
            paths.extend(folders)
    except botocore.exceptions.ClientError as e:
        if "AccessDenied" in repr(e):
            abort(f"You do not have access to {bucket_name}. Please review your AWS credentials")
    paths.sort(reverse=True)
    builds = []
    for p in paths[:num]:
        key = urljoin(p, buildinfo_filename)
        obj = s3.Object(bucket_name, key)
        f = BytesIO()
        try:
            obj.download_fileobj(f)
        except botocore.exceptions.ClientError:
            # most likely not found
            continue
        f.seek(0)
        c = json.load(f)
        c["s3_path"] = f"s3://{bucket_name}/{p}"
        builds.append(c)
    versions = {}
    builds.sort(key=itemgetter("version"), reverse=True)
    for build in builds:
        versions[build["version"]] = build
    return versions



import os
import os.path
import sys
import subprocess
import shutil
import pathlib
import webbrowser
from psutil import process_iter, NoSuchProcess, AccessDenied
import signal

from kamocli.ed.kamoplugin import KamoWorld, register, command


def kill_process_by_id(ports):
    # Start by killing the backend if its running
    for proc in process_iter():
        try:
            connections = proc.connections(kind="inet")
        except AccessDenied:
            continue  # Skipping processes that we don't have access to

        for con in connections:
            if con.laddr.port in ports:
                print(f"Killing Kamo backend on port: {con.laddr.port}")
                try:
                    proc.send_signal(signal.SIGTERM)
                except NoSuchProcess:
                    print("Process already killed")
                except AccessDenied:
                    print("Ignoring an access denied error")


@register
class LocalDev(KamoWorld):
    def __init__(self, ctx):
        super().__init__(ctx)

    @command
    def update_backend(self):
        backend_url = "git@github.com:kamo-io/kamo-platform.git"

        project_root = pathlib.Path(self.project_root).resolve().parent
        backend_root = project_root.joinpath("kamo-platform")

        print(project_root)
        if not backend_root.exists():

            # clone the project from git
            print("Cloning the backend for first time use")
            subprocess.call("git clone " + backend_url, cwd=project_root.as_posix(), shell=True)
            subprocess.call("pipenv install --dev", cwd=backend_root.as_posix(), shell=True)
            return

        else:
            print("Killing the backend if its already running")
            ports_to_kill = [7000, 7001, 7002]
            kill_process_by_id(ports_to_kill)

            print("Pulling latest...")
            subprocess.call("git pull", cwd=backend_root.as_posix(), shell=True)
            return

    @command
    def start_local_backend(self):
        ports_to_kill = [7000, 7001, 7002]
        kill_process_by_id(ports_to_kill)
        # This is hardcoded for demo's,  need to find a smarter way to start the backend through KamoED
        backend_root = pathlib.Path(self.project_root).resolve().parent.joinpath("kamo-platform")

        subprocess.Popen("start pipenv run kamoregistry", cwd=backend_root.as_posix(), shell=True)
        subprocess.Popen("start pipenv run kamorest", cwd=backend_root.as_posix(), shell=True)
        subprocess.Popen("start pipenv run knode jobs process -d", cwd=backend_root.as_posix(), shell=True)

    @command
    def open_kamoflage(self):
        """Open the kamo frontend locally"""
        webbrowser.open("https://app-vikingworld.kamo.io/", new=0, autoraise=True)

    @command
    def open_knode_web(self):
        """Open the kamo frontend locally"""
        webbrowser.open("https://app-vikingworld.kamo.io/", new=0, autoraise=True)

    @command
    def open_knode_logs(self):

        path = os.path.join(os.path.expanduser("~"), ".kamo", "knode_logs")

        if not os.path.exists(path):
            print(f"Path not found: {path}")
        else:
            if sys.platform == "win32":
                cmd = f'explorer /root, "{path}"'
            elif sys.platform == "darwin":
                cmd = ["open", "-R", path]
            else:
                print(f"Attempting to run on unsupported platform {sys.platform}")
                print(f"Point your file exploring application to: {path}")
                cmd = None

            if cmd:
                print(f"executing: {cmd}")
                subprocess.call(cmd)

    @command
    def open_local_db(self):

        path = os.path.join(os.path.expanduser("~"), ".kamo", self.project, "db")

        if not os.path.exists(path):
            print(f"Path not found: {path}")
        else:
            if sys.platform == "win32":
                cmd = f'explorer /root, "{path}"'
            elif sys.platform == "darwin":
                cmd = ["open", "-R", path]
            else:
                print(f"Attempting to run on unsupported platform {sys.platform}")
                print(f"Point your file exploring application to: {path}")
                cmd = None

            if cmd:
                print(f"executing: {cmd}")
                subprocess.call(cmd)

    @command
    def start_servers(self):
        """starts unreal dedicated server processes through the editor with uncooked content"""

        use_debug = self.ctx.obj.data["useDebug"]
        should_wait_for_debugger = self.ctx.obj.data["waitForDebugger"]
        overwrite_regions = self.ctx.obj.data["regionOverwrite"]

        log_cmds = self.ctx.obj.data["logCmds"]
        cmd_args = self.ctx.obj.data["cmdArgs"]
        cmd_args = [arg.strip() for arg in list(filter(None, cmd_args.split(",")))] + [f"-LogCmds={log_cmds}"]

        project_root = pathlib.Path(self.project_root).resolve()
        uproject_file = project_root.joinpath(self.project).as_posix() + ".uproject"

        engine_root = pathlib.Path(self.engine_root)

        if sys.platform == "win32":
            extention = ".exe"
            platform = "Win64"

        # Create the engine path
        executable_name = "ue4editor"

        if use_debug:
            executable_name = executable_name + "-Win64-DebugGame"

        if sys.platform == "darwin":
            if use_debug:
                executable_name = "UE4Editor-Mac-DebugGame.app/Contents/MacOS/UE4Editor-Mac-DebugGame"
            else:
                executable_name = "UE4Editor.app/Contents/MacOS/UE4Editor"

        editor_path = engine_root.joinpath("Binaries", platform, executable_name)

        if extention:
            editor_path = editor_path.as_posix() + extention

        # Get the region names from the database
        regions = []

        if overwrite_regions:
            regions = overwrite_regions.split(" ")
        else:
            session = self.get_kamotenant()
            db_path = self.db_url

            res = session.get(db_path, params={"object_type": "root"})
            res.raise_for_status()

            for each_root_object in res.json()["objects"]:
                region_name = each_root_object["id"]
                regions.append(region_name)

        base_port = 8000
        for i, region_name in enumerate(regions):
            port = str(base_port + i)

            cmd = [
                editor_path,
                uproject_file,
                "-log",
                "-server",
                f"-port={port}",
                "-KamoAllowCreateRegion",
                f"-regions={region_name}",
                f"LOG={region_name}.log",
                *cmd_args,
            ]

            if should_wait_for_debugger:
                cmd.append("-WaitForDebugger")

            subprocess.Popen(cmd)

    @command
    def delete_tenant(self):
        """delete world database"""
        path = os.path.join(os.path.expanduser("~"), ".kamo", self.project)

        if not os.path.exists(path):
            print(f"No tenant found: {path}")
            return

        shutil.rmtree(path)

    @command
    def restore_db_template(self):
        """restores the database from a template"""
        db_path = os.path.join(os.path.expanduser("~"), ".kamo", self.project, "db")

        # Hardcoded kamo ED path:
        kamoExtentions = f"{self.project_root}\KamoEdExtensions\db_templates"
        template_name = self.ctx.obj.data["template"]

        target_path = os.path.join(kamoExtentions, template_name)
        if not os.path.exists(target_path):
            print(f"Unable to find database template at :{target_path}")
            return

        # Delete the world database before applying the template
        self.delete_tenant()
        shutil.copytree(target_path, db_path)



import sys
import logging
import kamocli
import requests
import datetime
import dateutil
import json
import time
import copy
from time import sleep
import boto3
from urllib.parse import urlparse, urljoin
from io import BytesIO
from operator import attrgetter, itemgetter

from kamocli.utils import (
    dumps,
    abort,
    print_tab,
    fmt_date,
    check_tenant,
    print_table,
    get_tenant_url,
    success,
    find_regions,
    find_nodes,
    out,
    format_region_name,
    parse_region_name,
    get_builds_from_s3_url,
)
from kamosdk.exceptions import ServerError
from tabulate import tabulate

log = logging.getLogger(__name__)

from typer import Typer, Context, Argument, Option, echo, secho, confirm, style

app = Typer()


@app.callback(
    help="Manage jobs of UE4 server builds",
    context_settings={"allow_extra_args": True, "ignore_unknown_options": True},
)
@check_tenant
def cli(
    ctx: Context,
    is_all: bool = Option(False, "-a", "--all", is_flag=True, help="Manage all tenants at once"),
):
    ctx.obj.tenant = ctx.obj.session.profile.tenant
    ctx.obj.is_all = is_all


@app.command("list")
def list_jobs(
    ctx: Context,
    status: str = Option(
        "pending,processing,running,stopping",
        "-s",
        "--status",
        show_default=True,
        help="Status to filter for",
    ),
    is_all: bool = Option(False, "-a", "--all", is_flag=True, help="Show all jobs regardless of status"),
):
    """List server jobs"""
    tenant = ctx.obj.session.profile.tenant
    if ctx.obj.is_all:
        tenant = None
    if is_all:
        status = ""
    url = ctx.obj.session.knode_url + f"/jobs"
    resp = ctx.obj.session.get(
        url,
        params={
            "tenant": tenant,
            "start": 0,
            "end": 100,
            "status": status,
            "sort": "created",
            "reverse": True,
        },
    )
    if ctx.obj.output == "json":
        echo(dumps(resp.json()))
        return
    node_heartbeats = {r["node_id"]: None for r in resp.json()["jobs"]}
    for node_id in node_heartbeats.keys():
        url = ctx.obj.session.knode_url + f"/nodes/{node_id}"
        last_heartbeat = None
        try:
            node_resp = ctx.obj.session.get(url)
            last_heartbeat = node_resp.json()["last_modified"]
        except Exception:
            pass
        node_heartbeats[node_id] = last_heartbeat

    lst = []
    headers = [
        "ID",
        "Status",
        "Tenant",
        "Version",
        "Regions",
        "Node",
        "IP Address",
        "Created",
        "Updated",
        "Result",
        "Heartbeat",
    ]
    for job in resp.json()["jobs"]:
        args = job["args"]
        regions = ""
        for a in args:
            if a.startswith("-regions"):
                regions = a.split("=")[-1]
        address = "N/A"
        if job.get("ip_address"):
            address = "{}:{}".format(job.get("ip_address"), job.get("port"))
        heartbeat = node_heartbeats.get(job["node_id"])
        if heartbeat:
            dt = dateutil.parser.parse(heartbeat)
            heartbeat_txt = fmt_date(heartbeat)
            if dt < datetime.datetime.utcnow() - datetime.timedelta(minutes=5):
                heartbeat_txt = style(heartbeat_txt, fg="red")
        else:
            heartbeat_txt = style("Node not found", fg="red")
        lst.append(
            [
                job["job_id"],
                job["status"],
                job["tenant"],
                job["version"],
                regions,
                job["node_id"],
                address,
                fmt_date(job["created"]),
                fmt_date(job["last_modified"]),
                job.get("result"),
                heartbeat_txt,
            ]
        )
    if tenant:
        echo("Active jobs for tenant " + style(tenant, bold=True) + ":")
    else:
        echo("Active jobs for all tenants:")
    echo(tabulate(lst, headers=headers))


@app.command("view")
def view_job(ctx: Context, job_id: str = Argument(..., help="Job ID to view")):
    """View details about a specific job"""
    url = ctx.obj.session.knode_url + f"/jobs/{job_id}"
    try:
        resp = ctx.obj.session.get(url)
    except ServerError as ex:
        abort(ex)
    content = resp.json()
    if ctx.obj.output == "json":
        echo(dumps(content))
        return

    for key, val in content.items():
        if key == "args":
            continue
        elif key == "resources":
            txt = ""
            for k, v in val.items():
                txt += f"{k} = {v}, "
            if txt:
                txt = txt[:-2]
        else:
            txt = fmt_date(val)
        print_tab(key, txt)
    args = content["args"]
    echo("\nArguments:")
    for a in args:
        print_tab("", a)
    secho("\nNode status", bold=True)
    node_id = content["node_id"]
    url = ctx.obj.session.knode_url + f"/nodes/{node_id}"
    try:
        resp = ctx.obj.session.get(url)
    except ServerError:
        abort(f"Node {node_id} not found. This job is probably invalid.")
    node = resp.json()
    for k in [
        "node_id",
        "status",
        "last_modified",
        "registered",
        "instance_type",
        "availability_zone",
        "ami_id",
    ]:
        print_tab(k, node.get(k))
    resources = node.get("resource_allocations")
    if resources:
        print_tab("CPU Allocation", f"{resources['cpu']['current']}/{resources['cpu']['max']}")
        print_tab("RAM Allocation", f"{resources['ram']['current']}/{resources['ram']['max']}")
        print_tab("Utilization", f"{resources['utilization']}%")
    echo("")


@app.command("shutdown")
def shutdown_job(
    ctx: Context,
    job_id: str = Argument(..., help="Job ID to shut down"),
    is_force: bool = Option(False, "-y", is_flag=True, help="Force a shutdown without a prompt"),
):
    """Shut down the selected job by ID"""
    url = ctx.obj.session.knode_url + f"/jobs/{job_id}"
    try:
        resp = ctx.obj.session.get(url)
    except ServerError as ex:
        abort(ex)

    content = resp.json()
    job_id = content["job_id"]
    tenant = content["tenant"]
    if not is_force:
        confirm(
            f"Are you sure you want to shut down job '{job_id}' for the '{tenant}' tenant?",
            abort=True,
        )

    resp = ctx.obj.session.patch(url, json={"shutdown": True})

    if ctx.obj.output == "json":
        output = {"job": job_id, "tenant": tenant, "success": True, "command": "shutdown"}
        echo(json.dumps(output))
    else:
        secho(f"job '{job_id}' for tenant '{tenant}' is being shut down")


@app.command("delete")
def delete_job(
    ctx: Context,
    job_id: str = Argument(..., help="Job ID to shut down"),
    is_force: bool = Option(False, "-y", is_flag=True, help="Force a delete without a prompt"),
):
    """Delete the selected job by ID (dangerous!)"""
    url = ctx.obj.session.knode_url + f"/jobs/{job_id}"
    try:
        resp = ctx.obj.session.get(url)
    except ServerError as ex:
        abort(ex)

    content = resp.json()
    job_id = content["job_id"]
    tenant = content["tenant"]
    if not is_force:
        confirm(
            f"Are you sure you want to shut down job '{job_id}' for the '{tenant}' tenant?",
            abort=True,
        )

    resp = ctx.obj.session.delete(url)

    if ctx.obj.output == "json":
        output = {"job": job_id, "tenant": tenant, "success": True, "command": "shutdown"}
        echo(json.dumps(output))
    else:
        secho(f"job '{job_id}' for tenant '{tenant}' has been shut down")


@app.command("shutdown-all")
def shutdown_all(
    ctx: Context,
    is_force: bool = Option(False, "-y", is_flag=True, help="Force a shutdown without a prompt"),
    is_wait: bool = Option(False, "-w", "--wait", is_flag=True, help="Wait until jobs have been shut down"),
):
    """Shut down all jobs in the current tenant"""
    tenant = ctx.obj.tenant
    if not is_force:
        if ctx.obj.is_all:
            confirm(
                f"Are you sure you want to shut down ALL jobs for ALL tenants on the server?",
                abort=True,
            )
        else:
            confirm(
                f"Are you sure you want to shut down all jobs for the '{tenant}' tenant?",
                abort=True,
            )

    if ctx.obj.is_all:
        tenant = None
    url = ctx.obj.session.knode_url + f"/jobs"
    params = {"tenant": tenant, "start": 0, "end": 100, "status": "pending,processing,running"}
    resp = ctx.obj.session.get(
        url,
        params=params,
    )
    jobs = resp.json()["jobs"]
    cnt = len(jobs)
    for job in jobs:
        job_url = url + "/" + job["job_id"]
        _ = ctx.obj.session.patch(job_url, json={"shutdown": True})

    if is_wait:
        start_time = time.time()
        while time.time() < start_time + 30.0:
            resp = ctx.obj.session.get(
                url,
                params=params,
            )
            n = len(resp.json()["jobs"])
            if n <= 0:
                break
            out(f"Waiting for {n} jobs to stop")
            time.sleep(1.0)
        if n == 0:
            success(f"{cnt} jobs have been shut down for tenant '{tenant}'", bold=True)
        else:
            out(f"Timeout waiting for {n} jobs to stop", fg="red")
            out(
                f"It is possible that there are dead ec2 instances in which case you will need to delete jobs.\nYou can investigate this by running 'kamo nodes list' and then delete jobs with 'kamo jobs delete-all'."
            )

            sys.exit(1)
    else:
        if ctx.obj.is_all:
            success(
                f"{cnt} jobs are being shut down. All jobs on server have been sent the shutdown signal.",
                bold=True,
            )
        else:
            success(f"{cnt} jobs are being shut down for tenant '{tenant}'", bold=True)


@app.command("delete-all")
def delete_all(
    ctx: Context, is_force: bool = Option(False, "-y", is_flag=True, help="Force a delete without a prompt")
):
    """Delete all jobs in the current tenant (dangerous!)"""
    tenant = ctx.obj.tenant
    if not is_force:
        if ctx.obj.is_all:
            confirm(
                f"Are you sure you want to permanently delete ALL jobs for ALL tenants on the server?",
                abort=True,
            )
        else:
            confirm(
                f"Are you sure you want to permanently delete all jobs for the '{tenant}' tenant?",
                abort=True,
            )

    if ctx.obj.is_all:
        tenant = None
    url = ctx.obj.session.knode_url + f"/jobs"
    resp = ctx.obj.session.get(url, params={"tenant": tenant, "start": 0, "end": 100, "status": ""})
    for job in resp.json()["jobs"]:
        job_url = url + "/" + job["job_id"]
        _ = ctx.obj.session.delete(job_url)

    cnt = len(resp.json()["jobs"])
    if ctx.obj.is_all:
        success(
            f"{cnt} jobs have been deleted. All jobs on server have been removed.",
            bold=True,
        )
    else:
        success(f"{cnt} jobs have been deleted for tenant '{tenant}'", bold=True)


@app.command("cleanup")
def cleanup(ctx: Context):
    """Clean up orphaned jobs (with no node running) for the current tenant."""
    tenant = ctx.obj.tenant
    if ctx.obj.is_all:
        tenant = None
    jobs_url = ctx.obj.session.knode_url + f"/jobs"
    resp = ctx.obj.session.get(
        jobs_url, params={"tenant": tenant, "start": 0, "end": 100, "status": "running,pending,processing"}
    )
    n = 0
    for job in resp.json()["jobs"]:
        node_id = job["node_id"]
        url = ctx.obj.session.knode_url + f"/nodes/{node_id}"
        try:
            ctx.obj.session.get(url)
            continue
        except ServerError as e:
            if e.response["status_code"] == 404:
                out(f"Deleting orphaned job {job['job_id']} on node {job['node_id']} in tenant {job['tenant']}")
                job_url = jobs_url + "/" + job["job_id"]
                _ = ctx.obj.session.delete(job_url)
                n += 1
    if n:
        success(f"Done cleaning up {n} jobs", bold=True)
    else:
        success(f"No orphaned jobs to clean up", bold=True)


@app.command("create")
@check_tenant
def create_job(
    ctx: Context,
    s3_path: str = Option(None, "--s3", help="Deploy a specific build from S3"),
    version: str = Option(None, "-v", "--version", help="Version to deploy (or latest)"),
    cpu: int = Option(0, "-c", "--cpu", help="Amount of CPU's to reserve"),
    mem: str = Option("0 GB", "-m", "--mem", help="Amount of memory to reserve"),
    node_id: str = Option(None, "-n", "--node", help="Pick a specific node to run on"),
    volumes: str = Option(None, "--volumes", help="Level Volumes to load."),
    process_per_volume: bool = Option(
        False,
        "--ppv",
        is_flag=True,
        help="Deploy one process per Kamo Volume.",
    ),
    instance_id: str = Option(None, "-i", "--instance-id", help="Level instance ID."),
    executable: str = Option(None, "-e", "--executable", help="Executable name and path (if not standard)"),
    map_name: str = Option(
        None,
        "--map",
        help="Run a custom map or game mode, injected into the head of the command line",
    ),
    custom_args: str = Option("?listen,-server,-NullRHI", "-a", "-args", help="Override default arguments"),
    is_force: bool = Option(
        False,
        "-y",
        is_flag=True,
        help="Create the requested jobs without a prompt",
    ),
):
    """Create a new job. Deploy the most recent server build by default"""
    tenant = ctx.obj.tenant
    builds_path = ctx.obj.session.profile.builds_path
    if version and not builds_path:
        abort("You must set a build path with 'kamo builds path' to deploy a versioned build")

    if not version and not s3_path:
        abort("You must specify either a version (or latest) or s3_path")

    if process_per_volume:
        if not map_name and not volumes:
            abort(f"Map name or volumes must be specified when using --ppv option.")

        if map_name and not volumes:
            # PROVISIONAL: Enumerate regions, find the map, parse the name to get the volumes.
            map_name = map_name.lower()

            volumes_per_map = {}
            for region_name in find_regions(ctx):
                region_map_name, volume_name, region_instance_id = parse_region_name(region_name)
                print(f"Region: {region_map_name} {volume_name} {region_instance_id}")
                if instance_id != region_instance_id:  # We need exact match for the region
                    continue
                volumes_per_map.setdefault(region_map_name, []).append(volume_name)

            if map_name not in volumes_per_map:
                abort(
                    f"Map '{map_name}' not found on DB, only these exist: {list(volumes_per_map.keys())}.\nTry deploy without the --ppv switch to populate the DB."
                )

            volumes = " ".join([volume or "main" for volume in volumes_per_map[map_name]])

    args = [c.strip() for c in custom_args.split(",")]
    if not s3_path:
        versions = {}
        versions = get_builds_from_s3_url(builds_path)
        if not versions:
            abort(f"No builds found")
        if version != "latest":
            if version not in versions:
                abort(
                    f"Version '{version}' not found for tenant '{tenant}'.\nAvailable versions: {', '.join(versions.keys())}"
                )
        else:
            version = list(versions.keys())[0]
        s3_path = versions[version]["s3_path"]

    client = boto3.client("s3")
    parts = urlparse(s3_path)
    bucket_name = parts.netloc
    prefix = parts.path[1:]

    try:
        f = BytesIO()
        client.download_fileobj(bucket_name, prefix + "buildinfo.json", f)
        f.seek(0)
        build_info = json.load(f)
        if not version:
            version = build_info.get("version", "N/A")
        if not executable:
            executable = build_info.get("executable")
    except Exception as e:
        secho(f"Could not load buildinfo.json from {s3_path}. Was the path correct? Error: {e}", fg="red")

    if not executable:
        abort("You must specify an executable to run. e.fg. 'LinuxServer/TestProject/Binaries/Linux/TestProject'")

    all_nodes = find_nodes(ctx)
    if node_id:
        node_ids = all_nodes.keys()
        if node_id not in node_ids:
            abort(f"Node {node_id} not found")

    if instance_id:
        args.append(f"-kamoinstanceid={instance_id}")

    if volumes and not process_per_volume:
        # Add all the volumes on a single node
        args.append(f"-kamovolumes={volumes}")

    if process_per_volume:
        launches = volumes.split(" ")
    else:
        launches = [None]

    if not is_force:
        confirm(
            f"Do you want to create {len(launches)} jobs with version '{version}' from '{s3_path}' on the '{tenant}' tenant with executable '{executable}'?",
            abort=True,
        )

    responses = []
    for volume in launches:
        data = {
            "node_id": node_id,
            "tenant": tenant,
            "tenant_only": True,
            "version": version or "none",
            "map_name": map_name,
            "resources": {"cpu": cpu, "ram": mem},
            "executable": executable,
            "s3_path": s3_path,
            "args": copy.copy(args),
        }

        if process_per_volume:
            data["args"].append(f"-kamovolumes={volume}")

        log.info(
            f"Creating job with version {version} on 'region.{map_name}^{volume}#{instance_id}' for tenant {tenant}"
        )
        url = ctx.obj.session.knode_url + f"/jobs"
        try:
            resp = ctx.obj.session.post(url, json=data)
        except ServerError as ex:
            abort(ex)
        content = resp.json()
        responses.append(content)

    if ctx.obj.output == "json":
        echo(dumps(responses))
    else:
        for response in responses:
            job_id = response["job_id"]
            echo(f"job '{job_id}' created for tenant '{tenant}' with version {version}")


@app.command("maps")
def list_maps(
    ctx: Context,
    map_name_filter: str = Option(
        None,
        "--map",
        help="Only show this map",
    ),
    instance_filter: str = Option(
        None,
        "--instance",
        help="Only show this instance",
    ),
):
    """List maps available for jobs to deploy"""

    # PROVISIONAL: Enumerate regions, find the map, parse the name to get the volumes.
    if map_name_filter:
        map_name_filter = map_name_filter.lower()

    instances = {}
    for region_name in find_regions(ctx):
        map_name, volume_name, instance_id = parse_region_name(region_name)
        """
        print(f"Region: {region_map_name} {volume_name} {region_instance_id}")

        Region: map_main_p None None
        Region: map_worldglacier01_p volume_town01 None
        Region: map_worldglacier01_p None None
        """
        volume_name = volume_name or "main"
        instance_id = instance_id or "main"

        if map_name_filter and map_name != map_name_filter:
            continue
        if instance_filter and instance_filter != instance_id:
            continue

        instances.setdefault(instance_id, {}).setdefault(map_name, []).append(volume_name)

    if ctx.obj.output == "json":
        print(json.dumps(instances, indent=4))
    else:
        from textwrap import indent

        for instance_id, maps in instances.items():
            if len(instances) > 0:
                print(f"Instance: {instance_id}")
                fill = "    "
            else:
                fill = ""

            for map_name, volumes in maps.items():
                print(indent(f"{map_name}: {volumes}", fill))



"""
session
~~~~~~~~~~
Service Session, low-level object for communicating with RESTFul services.
"""

import requests
import requests.utils
from requests import codes
from os import environ
import copy
import json
import platform
import time
import logging
from hashlib import sha1
from typing import Optional, Union, Dict, List

from . import __version__
from .exceptions import ServerError, ServiceNotFound, InvalidProfile
from .utils import check_resp_error, get_access_token
from .config import Config, load_cache, save_cache, save_config

log = logging.getLogger(__name__)

config = Config()


class Profile:

    profile_name = None
    content = None

    def __init__(self, profile_name: Optional[str] = None, content: Optional[Dict] = None) -> None:
        if profile_name:
            self.profile_name = profile_name
            self.content = config.get("profiles")[self.profile_name]
        else:
            self.profile_name = None
            self.content = content

    def __getattr__(self, name):
        if name in self.content:
            return self.content[name]
        return None

    def __setattr__(self, name, value):
        if name in ("profile_name", "content"):
            super(Profile, self).__setattr__(name, value)
        else:
            self.content[name] = value
            if self.profile_name:
                save_config()


class ServiceSession(requests.Session):
    """
    A wrapped requests session object with base_url and endpoints
    from kamo service api's
    """

    def __init__(self, domain=None, api_key=None, profile=None, tenant=None, *args, **kwargs):
        super(ServiceSession, self).__init__(*args, **kwargs)
        # retry idempotent methods up to 5 times
        # if not environ.get("KAMO_DISABLE_RETRY"):
        #     retries = 5
        #     backoff_factor = 0.5
        #     status_forcelist = (500, 502, 503, 504)
        #     retry = Retry(
        #         total=retries,
        #         read=retries,
        #         connect=retries,
        #         backoff_factor=backoff_factor,
        #         status_forcelist=status_forcelist,
        #     )
        #     adapter = HTTPAdapter(max_retries=retry)
        #     self.mount("http://", adapter)
        #     self.mount("https://", adapter)

        # if no named profile or api key is passed in
        if not profile and not api_key:
            # find the default profile, if any
            if environ.get("KAMO_PROFILE"):
                profile = environ["KAMO_PROFILE"]
                log.info("Using profile '%s' from environment", profile)
            else:
                profile = config.get("default_profile")
                if profile:
                    log.info("Using default profile '%s' from config", profile)

        self.profile_name = profile

        # if we have a profile we will load it from the config
        if self.profile_name:
            log.info(
                "Initializing client with profile '%s'. Available profiles: %s",
                self.profile_name,
                list(config.get("profiles")),
            )
            if self.profile_name not in config.get("profiles"):
                raise InvalidProfile("The config profile (%s) could not be found" % profile)
            self.profile = Profile(self.profile_name)
        else:
            # if there is no profile there needs to be configuration set in the
            # environment, in which case we create an ephemeral profile, not
            # backed up by disk.
            api_key = api_key or environ.get("KAMO_API_KEY")
            if not api_key:
                raise InvalidProfile("No profile specified and KAMO_API_KEY not set in environment")
            domain = domain or environ.get("KAMO_DOMAIN")
            self.profile = Profile(content={"api_key": api_key, "domain": domain, "tenant": ""})

        self.root_info = {}
        self.endpoints = {}
        self.token = None

        if environ.get("KAMO_DOMAIN", domain):
            domain = environ.get("KAMO_DOMAIN", domain)
            log.info(f"Overriding domain with {domain}")
            self.profile.domain = domain
        if environ.get("KAMO_TENANT", tenant):
            tenant = environ.get("KAMO_TENANT", tenant)
            log.info(f"Overriding tenant with {tenant}")
            self.profile.tenant = tenant

        self.domain = self.profile.domain
        self.api_key = self.profile.api_key
        if domain:
            self.domain = domain
        if api_key:
            self.api_key = api_key

        if environ.get("KAMO_API_KEY"):
            log.info("Overriding api key from environment")
            self.api_key = environ["KAMO_API_KEY"]

        self.cache_name = sha1((self.domain or "").encode()).hexdigest()
        self.user_agent = "Kamo-SDK/%s Python/%s %s/%s" % (
            __version__,
            platform.python_version(),
            platform.system(),
            platform.release(),
        )

        self.headers = {
            "Content-Type": "application/json",
            "Accept": "application/json",
            "User-Agent": self.user_agent,
            "Kamo-Tenant": self.profile.tenant,
            "Kamo-Api-Key": self.api_key,
        }

        if not self._load():
            self._initialize()

        self.endpoints = self.root_info.get("endpoints")

    def _initialize(self) -> None:
        try:
            self.fetch_root_info()
        except ServerError:
            log.warning("Error fetching root info")


    def _save(self) -> None:
        contents = {
            "token": self.token,
            "root_info": self.root_info,
            "api_key": self.api_key,
        }
        save_cache(self.cache_name, contents)

    def _load(self) -> bool:
        contents = load_cache(self.cache_name)
        if not contents:
            return False
        self.token = contents["token"]
        self.root_info = contents["root_info"]
        if self.token:
            self.headers["Authorization"] = "Bearer {}".format(self.token)
        return True

    def _do_request(self, method, retry=True, *args, **kwargs):

        # method: GET
        old_content_type = self.headers["Content-Type"]
        if "headers" in kwargs:
            self.headers.update(kwargs["headers"])

        if method == "get":
            # ! Temporary hack: Remove the application/json content-type header for GET's
            del self.headers["Content-Type"]

        try:
            st = time.time()
            response = getattr(super(ServiceSession, self), method)(*args, **kwargs)
            diff = time.time() - st
        finally:
            self.headers["Content-Type"] = old_content_type

        # Manage response from the server
        log.info("%s %s returned %s in %.3f sec" % (method.upper(), args[0], response.status_code, diff))

        check_resp_error(response)
        return response

    def get_url(self, subdomain):
        return f"https://{subdomain}.{self.domain}"

    @property
    def kamo_url(self):
        return self.get_url("kamo")

    @property
    def knode_url(self):
        return self.get_url("knode")

    def get(self, *args, **kw):
        return self._do_request("get", True, *args, **kw)

    def put(self, *args, **kw):
        return self._do_request("put", True, *args, **kw)

    def post(self, *args, **kw):
        return self._do_request("post", True, *args, **kw)

    def delete(self, *args, **kw):
        return self._do_request("delete", True, *args, **kw)

    def fetch_root_info(self) -> Dict:
        log.debug("fetch_root_info(): domain: {0}, headers: {1}".format(self.domain, self.headers))
        root_url = self.get_url("kamo")
        try:
            r = requests.get(root_url, timeout=3.0, headers=self.headers)
        except requests.exceptions.ConnectionError as ex:
            raise ServerError(f"Could not reach server {root_url} ({ex})") from None
        except requests.exceptions.ReadTimeout as ex:
            raise ServerError(f"Could not reach server {root_url} ({ex})") from None

        r.raise_for_status()

        if "application/json" not in r.headers["Content-Type"]:
            raise ServerError("Unexpected response: %s" % r.text, url=root_url) from None
        ret = r.json()
        self.root_info = ret
        return ret

    def url_from_endpoint(self, endpoint: str) -> str:
        try:
            return self.endpoints[endpoint]
        except KeyError:
            raise ServerError(
                "Endpoint '%s' is not exported by '%s'.\nAvailable endpoints are %s"
                % (endpoint, self.root_url, ", ".join(self.endpoints.keys()))
            )

    def request(self, method, url, **kwargs):
        log.debug("Calling %s %s %s", method, url, repr(kwargs))
        stripped_headers = copy.copy(self.headers)
        if "Authorization" in stripped_headers:
            stripped_headers["Authorization"] = "Bearer ***"
        log.debug("Headers: %s" % json.dumps(stripped_headers))
        if "json" in kwargs:
            log.debug("Payload:\n%s" % json.dumps(kwargs["json"], indent=4))
        return super(ServiceSession, self).request(method, url, **kwargs)

    def links(self, resp: Dict) -> Dict:
        return resp.get("links", {})



import os
import requests
from urllib.parse import urlsplit, urljoin

from kamosdk.config import (
    root_config_folder,
    get_profiles,
    get_default_profile,
    set_default_profile,
    create_profile,
    delete_profile,
    clear_config,
)
from kamosdk.exceptions import InvalidProfile
from kamocli.utils import abort, dumps

from typer import Typer, Context, Argument, Option, echo, secho, confirm, prompt

app = Typer()


@app.command("list")
def list_profiles(ctx: Context):
    """lists all available profiles."""
    if ctx.obj.output == "json":
        echo(dumps(get_profiles()))
        return
    echo("Available profiles:")
    profiles = get_profiles()
    for profile_name in sorted(profiles):
        domain = profiles[profile_name]["domain"]
        col = ""
        if get_default_profile() == profile_name:
            col = "green"
            domain += " (selected)"
        secho(f"{profile_name:10} > {domain}", fg=col)
    echo("")


@app.command("current")
def current(ctx: Context):
    """displays content for the profile name."""
    if not get_profiles():
        abort("No profiles are currently set up. Please run 'kamo profile add [name]' to start")

    profile = ctx.obj.session.profile_name
    secho("Current profile: %s" % get_default_profile(), fg="green")
    profile_config = get_profiles().get(profile)
    if not profile_config:
        secho("Profile %s not found" % profile, fg="red")
        ctx.invoke(list)
        return

    for k, v in sorted(profile_config.items()):
        echo("   {0:20}{1}".format(k, v))
    echo("")


@app.command("view")
def view(ctx: Context, profile: str):
    """displays content for a given profile name."""
    if not get_profiles():
        abort("No profiles are currently set up. Please run 'kamo profile add [name]' to start")

    profile = (profile or "").lower()
    secho("Current profile: %s" % get_default_profile(), fg="green")
    profile_config = get_profiles().get(profile)
    if not profile_config:
        secho("Profile %s not found" % profile, fg="red")
        ctx.invoke(list)
        return

    echo("Viewing profile configuration for %s" % profile)
    for k, v in sorted(profile_config.items()):
        echo("   {0:20}{1}".format(k, v))
    echo("")


@app.command("set")
def set_profile(ctx: Context, profile: str):
    """sets the profile given by name as 'current'."""
    profile = profile.lower()
    try:
        set_default_profile(profile)
    except Exception as e:
        secho(str(e), fg="red")
        echo("View available profiles with 'kamo profile list'")
        return
    else:
        secho("Your profile has been changed to %s" % profile, bold=True)


@app.command("edit")
def edit_profile(ctx: Context):
    """open up the config.yaml file"""
    full_filename = os.path.join(root_config_folder, "config.yaml")
    echo("Editing config file located at %s" % (full_filename))

    echo("Modify your config and close the file")
    from click import edit

    edit(filename=full_filename, require_save=True)


@app.command("add")
def add_profile(
    ctx: Context,
    profile: str = Option(..., "-p", "--profile", prompt="What would you like to call this profile?"),
    domain: str = Option(..., "-d", "--domain", prompt="Domain name of the service (e.g. kamo.io)"),
    api_key: str = Option(..., "-k", "--api-key", prompt="API Key"),
    replace: bool = Option(False, "-r", "--replace", is_flag=True, help="Replace an existing profile"),
    tenant: str = Option(None, "-t", "--tenant", help="Default tenant (e.g. staging)"),
    skip_auth: str = Option(
        False, "-s", "--skip-auth", is_flag=True, help="Skip authentication (for local development)"
    ),
):
    """Creates a new profile of the specified name and allows you to configure it"""
    profile = profile.lower()
    if profile in get_profiles() and not replace:
        secho(
            "Profile '{}' already exists. Use --replace if you want to override it.".format(profile),
            fg="red",
        )
        return

    if "/" in domain:
        secho("Domain should only look like this: domain.io", fg="red")
        return

    if not skip_auth:
        url = f"https://kamo.{domain}/.tenants"
        try:
            resp = requests.get(url, timeout=2, headers={"Kamo-Api-Key": api_key})
            while resp.status_code == 403:
                api_key = prompt("API Key incorrect. Please try again")
                url = f"https://kamo.{domain}/.tenants"
                resp = requests.get(url, timeout=2, headers={"Kamo-Api-Key": api_key})

            resp.raise_for_status()
        except Exception as e:
            secho(f"{url} is not reachable: {e}", fg="red")
            if not confirm("Are you sure you want to add this domain?"):
                return

    try:
        create_profile(profile, api_key=api_key, domain=domain, skip_auth=skip_auth, tenant=tenant)
    except InvalidProfile as ex:
        abort(ex)
    set_default_profile(profile)
    echo("Profile has been added and is now active. If you want to edit it please run: kamo profile edit.")
    if not tenant:
        secho("Please select a tenant with 'kamo tenants list' / 'kamo tenants set'")


@app.command("delete")
def delete(ctx: Context, profile: str = Argument(..., help="Profile name to delete")):
    """accepts a profile name and deletes it."""
    profile = profile.lower()
    delete_profile(profile)
    secho("Profile '%s' has been deleted" % profile, bold=True)


@app.command("clear")
def delete(ctx: Context):
    """delete all profiles"""
    clear_config()
    secho("All profiles have been deleted and cache has been cleared. Please use 'kamo profile add' to get started", bold=True)



"""
KamoEd - Kamo Editor CLI Tool

A Command Line Interface for dispatching KamoEd plugin commands.
"""
import sys
import os.path
import traceback
import tempfile
import json
from functools import wraps
from contextlib import contextmanager

import click
import requests

from kamocli.ed import kamoplugin


class JsonResponser(click.Group):
    """
    Turn return values from commands into json formatted response.
    Catch any exceptions and pretty print as json as well.
    """

    def invoke(self, ctx):
        ctx.obj.json = {}
        ctx.obj.json["_command"] = sys.argv
        ctx.obj.json["success"] = False

        try:
            ret = super(JsonResponser, self).invoke(ctx)
            if ret:
                ctx.obj.json.update(ret)
            ctx.obj.json["success"] = True

        except click.exceptions.Exit as e:
            pass
        except requests.exceptions.HTTPError as e:
            ctx.obj.json = {
                "status_code": e.response.status_code,
                "error": e.response.json().get("error", e.response.text),
                "request": "{} {}".format(e.request.method, e.request.url),
            }
            traceback.print_exc()
        finally:
            if getattr(ctx.obj, "stdout", None):
                print("----BEGIN JSON-----")
                print(json.dumps(ctx.obj.json, indent=4))
                print("----END JSON-----")
            else:
                with tempfile.TemporaryFile(mode="w", delete=False) as f:
                    print("JSON result file: {}".format(f.name))
                    json.dump(ctx.obj.json, f, indent=4)


@click.group(cls=JsonResponser, context_settings={"help_option_names": ["-h", "--help"]})
@click.option("--project", help="Name of the project")
@click.option("--project-root", help="Project root path")
@click.option("--engine-root", help="Engine root path")
@click.option("--plugin", help="Name of plugin class. Default is 'kamoworld'")
@click.option("--data", type=click.File(mode="r"), help="Path to input data json file.")
@click.option(
    "--include",
    type=click.Path(exists=True, file_okay=False, dir_okay=True, readable=True),
    multiple=True,
    help="Path to plugins folder. Multiple include options are supported.",
)
@click.option("--stdout", is_flag=True, help="Write result to stdout instead of tempfile.")
@click.pass_context
def cli(ctx, project, project_root, engine_root, plugin, data, include, stdout):
    """KamoEd - Unreal Engine Kamo Editor Extension."""
    print(ctx.obj.base_path)
    ctx.obj.project = project
    ctx.obj.project_root = project_root
    ctx.obj.engine_root = engine_root
    ctx.obj.plugin = plugin
    ctx.obj.include = include
    ctx.obj.stdout = stdout

    if data:
        ctx.obj.data = json.load(data)
    else:
        ctx.obj.data = {}

    path = ctx.obj.base_path / "ed" / "coreplugins"
    kamoplugin.load_plugins(path)

    for path in ctx.obj.include:
        kamoplugin.load_plugins(path)


@cli.command()
@click.pass_context
def info(ctx):
    """Get information on this command line tool."""
    return {"plugins": list(kamoplugin.plugins)}


@cli.command()
@click.pass_context
@click.option("--auth_url", help="OAuth2 endpoint for authorization.")
@click.option("--auth_token", help="OAuth2 token for authorization.")
@click.option("--kamo_registry_url", help="KamoRegistry endpoint.")
@click.option("--kamo_url", help="KamoREST endpoint.")
@click.option("--kamo_token", help="KamoRegistry/REST authorization token.")
@click.option("--tenant", help="KamoREST tenant to use.", default="default")
@click.option("--docker-local-server", help="Build and use a local image for the server.")
@click.argument("command")
def wc(
    ctx,
    auth_url,
    auth_token,
    kamo_registry_url,
    kamo_url,
    kamo_token,
    tenant,
    docker_local_server,
    command,
):
    """
    Execute a world command.

    Run 'list-commands' to get a list of available commands.

    An instance of KamoWorld is created and a command is executed on the world.
    The result is formatted as a json document and written to a temporary file. The name of the file
    is displayed in the output.
    """

    def trailing_slash(url):
        if not url or url.endswith("/"):
            return url
        else:
            return url + "/"

    ctx.obj.auth_url = auth_url
    ctx.obj.auth_token = auth_token
    ctx.obj.kamo_registry_url = trailing_slash(kamo_registry_url)
    ctx.obj.kamo_url = trailing_slash(kamo_url)
    ctx.obj.kamo_token = kamo_token
    ctx.obj.tenant = tenant
    ctx.obj.docker_local_server = docker_local_server

    plugin = ctx.obj.plugin or "kamoworld"  # The default plugin
    if plugin not in kamoplugin.plugins:
        raise RuntimeError("No plugin named '{}' found! Plugins: {}".format(plugin, list(kamoplugin.plugins)))

    world_class = kamoplugin.plugins[plugin]
    world = world_class(ctx)

    # Assign user values
    world.auth_url = ctx.obj.auth_url
    world.auth_token = ctx.obj.auth_token
    world.kamo_url = ctx.obj.kamo_url
    world.kamo_token = ctx.obj.kamo_token
    world.tenant_name = ctx.obj.tenant
    world.docker_local_server = True

    # Assign session values
    world.command_args = ctx.obj.data

    print(">> Executing '{}' in plugin {}".format(command, world.__class__))

    ret = world.execute_command(command)

    return ret



import os
import sys

import requests
import logging
import http.server
import socketserver
import threading

from urllib.parse import urlparse, parse_qs

from kamosdk.config import get_profile_config, get_default_profile, create_profile
from kamocli.utils import abort, dumps

log = logging.getLogger(__name__)

from typer import Context, launch, echo, secho
from kamocli.__main__ import app


@app.command("login", help="Authenticate")
def cli(ctx: Context):
    raise NotImplementedError(
        "This functionality is currently not implemented. Please use 'kamo auth' and enter API Key"
    )
    PORT = 9898
    redirect_uri = f"http://localhost:{PORT}"

    profile = get_profile_config()
    root_info = ctx.obj.session.root_info

    client_id = root_info["config"]["auth_client_id"]
    cognito_url = root_info["config"]["auth_url"]
    url = f"{cognito_url}/login?client_id={client_id}&response_type=code&scope=email+openid+profile&redirect_uri={redirect_uri}"
    launch(url)

    httpd = None

    class Handler(http.server.SimpleHTTPRequestHandler):
        def do_GET(self):
            self.send_response(200)
            self.send_header("Content-type", "text/html")
            self.end_headers()
            query_components = parse_qs(urlparse(self.path).query)

            html = """
                <html>
                <head>
                <script>
                setTimeout("window.close();", 100);
                </script>
                </head>
                <body>
                <h1>You have been logged in</h1>
                <p>You can now close this window.</p>
                </body>
                </html>
                """

            self.wfile.write(bytes(html, "utf8"))

            # if we have a proper request, finish the song-and-dance
            # against keycloak and create the profile with the returned
            # refresh token (api key)
            if query_components:
                headers = {"Content-Type": "application/x-www-form-urlencoded"}
                data = {
                    "grant_type": "authorization_code",
                    "code": query_components["code"],
                    "client_id": client_id,
                    "redirect_uri": redirect_uri,
                }
                url = f"{cognito_url}/oauth2/token"
                resp = requests.post(
                    url,
                    data=data,
                    headers=headers,
                )
                refresh_token = resp.json()["refresh_token"]
                create_profile(get_default_profile(), api_key=refresh_token, domain=profile["domain"])
                secho("You have been logged in", bold=True)

                # shut down the webserver in another thread to avoid deadlock
                threading.Thread(target=httpd.shutdown).start()

            return

    echo("Waiting for webpage...")
    socketserver.TCPServer.allow_reuse_address = True
    httpd = socketserver.TCPServer(("", PORT), Handler)
    httpd.serve_forever()
    # Authentication flow will be completed in do_GET()



#!/usr/bin/env python
import os, sys
from setuptools import setup, find_packages, Command

if sys.version_info < (3, 6):
    raise NotImplementedError("""This tool does not support Python versions older than 3.6""")


version_file = os.path.abspath(os.path.join("kamocli", "VERSION"))


def get_version():
    with open(version_file) as f:
        return f.readlines()[0].strip()


class BumpCommand(Command):
    description = "Version bump"
    user_options = [("which=", None, "Specify which part of the version to bump")]

    def initialize_options(self):
        self.which = None

    def finalize_options(self):
        assert self.which in (None, "major", "minor", "patch"), "Invalid bump pragma"

    def run(self):
        bump_version(self.which)


def bump_version(which):
    major, minor, patch = (0, 1, 0)
    old_version = get_version()
    major, minor, patch = old_version.split(".")

    if which == "major":
        major = int(major) + 1
        minor = 0
        patch = 0
    elif which == "minor":
        minor = int(minor) + 1
        patch = 0
    else:
        patch = int(patch.split("-")[0]) + 1

    new_version = "%s.%s.%s" % (major, minor, patch)

    with open(version_file, "w") as f:
        f.write(new_version)
    print(f"Version has been bumped from {old_version} to {new_version}")


with open("README.md", "r") as fh:
    long_description = fh.read()

version = get_version()

setup(
    name="kamo-cli",
    python_requires=">=3.6",
    version=version,
    license="MIT",
    author="Directive Games",
    author_email="info@directivegames.com",
    description="Kamo Commandline Tool",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="",
    packages=find_packages(exclude=["tests"]),
    package_data={"": ["VERSION"]},
    install_requires=[
        "typer",
        "click",
        "requests",
        "boto3",
        "psutil",
        "jsonpath-rw",
        "pyyaml",
        "pyjwt",
        "tabulate",
        "colorama",
        "shellingham",
        "boto3",
        "loguru",
    ],
    extras_require={},
    include_package_data=False,
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
    ],
    test_suite="tests",
    cmdclass={"bump": BumpCommand},
    entry_points={"console_scripts": ["kamo = kamocli.__main__:app"]},
)



import sys
import os
import os.path
import traceback
from importlib import import_module

import requests

plugins = {}
commands = []


def load_plugins(path):
    """Load plugins by importing script modules from 'path'."""
    sys.path.insert(0, str(path))
    for modulename in os.listdir(path):
        if modulename.lower().endswith(".py") and not modulename.startswith("__"):
            pure_modulename, ext = os.path.splitext(modulename)
            print("Loading in plugin: {}".format(os.path.join(path, modulename)))
            try:
                _ = import_module(pure_modulename)
            except Exception:
                print("Error: Can't import module: {}".format(os.path.join(path, modulename)))
                traceback.print_exc()
                raise


def register(c):
    print("Registering class '{}' from module {}".format(c.__name__, c.__module__))
    plugins[c.get_name()] = c
    return c


def command(f):
    commands.append(f.__name__.replace("_", "-"))
    return f


@register
class KamoWorld(object):
    def __init__(self, ctx):
        self.ctx = ctx
        self.project: str = ctx.obj.project
        self.project_root: str = ctx.obj.project_root
        self.engine_root: str = ctx.obj.engine_root

        # Initialized by user
        self.auth_url: str = None
        self.auth_token: str = None
        self.kamo_url: str = None
        self.kamo_token: str = None
        self.tenant_name: str = None

        # Session variables
        self.session = None
        self.command_args = {}
        self.tenant = None
        self.db_url: str = None
        self.mq_url: str = None

    @classmethod
    def get_name(c):
        return c.__name__.lower()

    def get_kamorest(self):
        """Return a requests.Session object to a KamoREST endpoint. Do authentication if needed."""
        if self.session:
            return self.session

        session = requests.Session()
        session.timeout = 2.0
        self.kamo_url = self.kamo_url or "http://127.0.0.1:7000/"
        self.db_url = self.kamo_url + "db"
        self.mq_url = self.kamo_url + "mq"

        if not self.kamo_token and all([self.auth_url, self.auth_token]):
            ret = session.post(
                self.auth_url,
                data="grant_type=client_credentials",
                headers={
                    "Authorization": "Basic " + auth_token,
                    "Accept": "application/json",
                    "Content-Type": "application/x-www-form-urlencoded",
                },
            )

            ret.raise_for_status()
            resp = ret.json()

            self.kamo_token = resp["access_token"]
            self.token_expires_at_utc = utcnow() + resp["expires_in"]

        if self.kamo_token:
            session.auth = "Bearer " + self.kamo_token

        if self.tenant_name:
            session.headers["Kamo-Tenant"] = self.tenant_name

        self.session = session
        return self.session

    def get_kamotenant(self):
        if not self.tenant_name:
            raise RuntimeError("Tenant name must be specified!")
        self.get_kamorest()

        ret = self.session.get(self.kamo_url)
        ret.raise_for_status()
        self.tenant = ret.json()

        if not self.tenant:
            raise RuntimeError("Tenant {} not found at {}.".format(self.tenant_name, self.kamo_url))

        return self.session

    def _get_command(self, command):
        """Get member function that matches 'command' with dashes and whitespaces converted to underscore."""
        command = command.replace("-", "_").replace(" ", "_")
        fn = getattr(self, command, None)
        return fn

    def execute_command(self, command):
        fn = self._get_command(command)
        if not fn:
            raise RuntimeError("Command not found: {}".format(command))
        return fn()

    @command
    def list_commands(self):
        """List executable commands."""
        ret = []
        for command in commands:
            fn = self._get_command(command)
            if fn:
                entry = {}
                entry["name"] = command
                if fn.__doc__:
                    entry["description"] = fn.__doc__
                ret.append(entry)

        return {"commands": ret}

    @command
    def get_registry(self):
        raise RuntimeError("Not implemented yet")

    @command
    def get_world(self):
        """Returns all world info in a single json document."""
        s = self.get_kamotenant()

        resp = {}
        resp["tenant"] = self.tenant

        # Get all objects and group them by type
        groups = {}
        ret = s.get(self.db_url)
        ret.raise_for_status()
        for ob in ret.json()["objects"]:
            groups.setdefault(ob["object_type"], []).append(ob)
        resp["object_groups"] = groups

        return resp

    @command
    def create_tenant(self):
        """Create a new tenant. Specify 'tenant_name'."""
        if "tenant_name" not in self.command_args:
            raise RuntimeError("In create_tenant: 'tenant_name' must be specified in command args.")
        s = self.get_kamorest()
        ret = s.post(
            self.kamo_url + "api/",
            json={"name": self.command_args["tenant_name"], "db": "file", "mq": "file"},
        )
        ret.raise_for_status()
        return ret.json()
